'Copyright (c) 2014 Donald Milne
'All rights reserved.
'
'This file is part of AOSkills4.
'
'AOSkills4 is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'AOSkills4 is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with AOSkills4. If not, see <http://www.gnu.org/licenses/>.

Imports System
Imports System.Windows.Forms
Imports System.Drawing

Friend Class aosEquipSlot

    Private prv_Slot As Integer
    Private prv_Page As Integer
    Private prv_LO_AOID As Integer
    Private prv_HI_AOID As Integer
    Private prv_QL As Integer

    Private Shared prv_page_name() As String = {"Weapon", "Armor"}

    Private Shared prv_slot_name(,) As String = { _
      {"HUD1", "HUD2", "HUD3", "UTIL1", "UTIL2", "UTIL3", "WEAPON_R", "BELT", "WEAPON_L", "NCU1", "NCU2", "NCU3", "NCU4", "NCU5", "NCU6"}, _
      {"NECK", "HEAD", "BACK", "SHOULDER_R", "BODY", "SHOULDER_L", "ARM_R", "HANDS", "ARM_L", "WRIST_R", "LEGS", "WRIST_L", "FINGER_R", "FEET", "FINGER_L"} _
    }

    Private Shared prv_ao_slot_id(,) As Integer = { _
     {2, 32768, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384}, _
     {2, 4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192, 16384, 32768} _
    }

    Private Shared prv_SelectedEquipSlot As aosEquipSlot = Nothing

    Public ReadOnly Property Slot() As Integer
        Get
            Return prv_Slot
        End Get
    End Property

    Public ReadOnly Property Page() As Integer
        Get
            Return prv_Page
        End Get
    End Property

    Public Property LO_AOID() As Integer
        Get
            Return prv_LO_AOID
        End Get
        Set(ByVal Value As Integer)
            prv_LO_AOID = Value
        End Set
    End Property

    Public Property HI_AOID() As Integer
        Get
            Return prv_HI_AOID
        End Get
        Set(ByVal Value As Integer)
            prv_HI_AOID = Value
        End Set
    End Property

    Public Property QL() As Integer
        Get
            Return prv_QL
        End Get
        Set(ByVal Value As Integer)
            prv_QL = Value
        End Set
    End Property

    ' return the picturebox instance
    Public ReadOnly Property PB() As PictureBox
        Get
            Return Me.pbSlot
        End Get
    End Property

    ' returns Weapon or Armor
    Public ReadOnly Property PageName() As String
        Get
            Return prv_page_name(prv_Page - 1)
        End Get
    End Property

    ' returns the short slot name (e.g. HUD1)
    Public ReadOnly Property SlotName() As String
        Get
            Return prv_slot_name(prv_Page - 1, prv_Slot - 1)
        End Get
    End Property

    ' returns the AO slot id for queries
    Public ReadOnly Property SlotID() As Integer
        Get
            Return prv_ao_slot_id(prv_Page - 1, prv_Slot - 1)
        End Get
    End Property

    ' returns the currently selected instance
    Public Shared ReadOnly Property SelectedEquipSlot() As aosEquipSlot
        Get
            Return prv_SelectedEquipSlot
        End Get
    End Property

    ' convert page and slot into a single value (in range 1 to 30) - will map to the position in storage array (or position +1 as array will be 0 to 29)
    Public ReadOnly Property SelectedSlot() As Integer
        Get
            Return 15 * (prv_Page - 1) + prv_Slot
        End Get
    End Property

    ' sets the bitmap of the picturebox and ensures transparency set correctly
    Public Sub SetImage(ByRef img As System.Drawing.Bitmap)
        Me.pbSlot.Image = img
        DirectCast(Me.pbSlot.Image, System.Drawing.Bitmap).MakeTransparent(Color.Lime)
    End Sub

    ' populates a modifiers array if there is an item attached to this slot
    Public Sub GetModifiers(ByRef tmpMods() As Integer)
        If prv_QL > 0 Then
            getItemModifiers(prv_LO_AOID, prv_HI_AOID, prv_QL, tmpMods)
        End If
    End Sub

    ' attaches an item to this slot and sets the icon and tooltip
    Public Sub SetItem(ByVal lo_aoid As Integer, ByVal hi_aoid As Integer, ByVal ql As Integer)
        Dim s_icon As String
        Dim icon As Bitmap

        Me.LO_AOID = lo_aoid
        Me.HI_AOID = hi_aoid
        Me.QL = ql
        If lo_aoid > 0 And Not (fmPDoll Is Nothing) Then
            ' get the icon for this
            s_icon = getIcon(Me.LO_AOID).ToString()
            Try
                icon = New System.Drawing.Bitmap(home_dir & "\icons\" & s_icon & ".png")
            Catch ex As System.ArgumentException
                ' our icon doesn't exist... use a generic one
                icon = New System.Drawing.Bitmap(home_dir & "\default.png")
            End Try
            Me.SetImage(icon)
            fmPDoll.ttECForm.SetToolTip(Me.PB, getName(Me.HI_AOID) & " @ QL " & Me.QL.ToString().Trim())
        ElseIf Not (fmPDoll Is Nothing) Then
            Me.pbSlot.Image = Nothing
            fmPDoll.ttECForm.SetToolTip(Me.PB, "")
        End If
    End Sub

    ' just an alias
    Public Sub ClearItem()
        Me.SetItem(0, 0, 0)
    End Sub

    Private Sub pbSlot_Click(sender As System.Object, e As System.EventArgs) Handles pbSlot.Click
        ' point the shared property at this instance
        prv_SelectedEquipSlot = Me
        fmPDoll.populateListBox(Me.PageName, Me.SlotID)
        fmPDoll.findInList()
        fmPDoll.populateUserModifierList()
    End Sub

    Public Sub New(ByVal page As Integer, ByVal slot As Integer)
        ' This call is required by the designer.
        InitializeComponent()
        ' Add any initialization after the InitializeComponent() call.
        prv_Page = page
        prv_Slot = slot
        Me.ClearItem()
    End Sub

End Class
