'Copyright (c) 2005-2014 Donald Milne
'All rights reserved.
'
'This file is part of AOSkills4.
'
'AOSkills4 is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'AOSkills4 is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with AOSkills4. If not, see <http://www.gnu.org/licenses/>.

Imports System
Imports System.Drawing
Imports System.Collections
Imports System.Windows.Forms
Imports Microsoft.VisualBasic

Public Class aosImplantItem : Inherits UserControl

    Private abilityAbrev() As String = {"Str", "Agl", "Sta", "Int", "Sen", "Psy"}
    Private abilityFull() As String = {"Strength", "Agility", "Stamina", "Intelligence", "Sense", "Psychic"}

    Private prvItemIdx As Integer = -1
    Private prvToolTip As ToolTip
    Private prvHandler As EventHandler
    Private prvWindow As System.Windows.Forms.Form

    Private prvQL As Integer = 1
    Private prvMaxQL As Integer = 1
    Private prvProfID As Integer = -1
    Private prvBreedID As Integer = -1

    Private prvUseHandler As Boolean = False
    Private prvCheckReqs As Boolean = True

    Private prvMods(NUM_SKILLS + NUM_ATTRS + NUM_OTHER_STATS + 1) As Integer

    Private prvOldType As String
    Private prvOldName As String
    Private prvOldQL As Integer
    Private prvOldShiny As String
    Private prvOldBright As String
    Private prvOldFaded As String

    Public Property UseHandler() As Boolean
        Get
            Return prvUseHandler
        End Get
        Set(ByVal Value As Boolean)
            prvUseHandler = Value
        End Set
    End Property

    Public Property CheckReqs() As Boolean
        Get
            Return prvCheckReqs
        End Get
        Set(ByVal Value As Boolean)
            prvCheckReqs = Value
        End Set
    End Property

    Public Property ItemIdx() As Integer
        Get
            Return prvItemIdx
        End Get
        Set(ByVal Value As Integer)
            prvItemIdx = Value
        End Set
    End Property

    Public Property QL() As Integer
        Get
            Return prvQL
        End Get
        Set(ByVal Value As Integer)
            prvQL = Value
            Me.nudQL.Value = Value
            Me.lblQL.Text = CStr(Value)
        End Set
    End Property

    Public Property Lock() As Boolean
        Get
            Return Me.chkLock.Checked
        End Get
        Set(ByVal Value As Boolean)
            Me.chkLock.Checked = Value
            If Value Then
                Me.chkLock.Text = "X"
            Else
                Me.chkLock.Text = "O"
            End If
        End Set
    End Property

    Public Property MaxQL() As Integer
        Get
            Return prvMaxQL
        End Get
        Set(ByVal Value As Integer)
            If Value < 1 Then
                prvMaxQL = 1
            ElseIf Value > 300 Then
                prvMaxQL = 300
            Else
                prvMaxQL = Value
            End If
            If Me.nudQL.Value > prvMaxQL Then
                Me.nudQL.Value = prvMaxQL
                Me.lblQL.Text = CStr(prvMaxQL)
            End If
            Me.nudQL.Maximum = New Decimal(New Integer() {prvMaxQL, 0, 0, 0})
        End Set
    End Property

    Public Property ProfessionID() As Integer
        Get
            Return prvProfID
        End Get
        Set(ByVal Value As Integer)
            prvProfID = Value
            If Value = SHADE_ID Then
                If Me.lblType.Text = "A" Then
                    changeLabel()
                End If
                If Me.lblType.Text = "I" Then
                    changeLabel()
                End If
            End If
            If Me.lblType.Text = "S" Then
                setSymbsAndSpirits() ' currently assumes false and false as params, we should track the form's chkReqs flag somehow?
            End If
        End Set
    End Property

    Public Property BreedID() As Integer
        Get
            Return prvBreedID
        End Get
        Set(ByVal Value As Integer)
            prvBreedID = Value
        End Set
    End Property

    Public Property Window() As System.Windows.Forms.Form
        Get
            Return prvWindow
        End Get
        Set(ByVal Value As System.Windows.Forms.Form)
            prvWindow = Value
        End Set
    End Property

    Public ReadOnly Property ShinyCluster() As String
        Get
            If Me.cmbShiny.SelectedIndex <> -1 Then
                Return Me.cmbShiny.Items(Me.cmbShiny.SelectedIndex).ToString()
            Else
                Return SEL_STRING
            End If
        End Get
    End Property

    Public ReadOnly Property BrightCluster() As String
        Get
            If Me.cmbBright.SelectedIndex <> -1 Then
                Return Me.cmbBright.Items(Me.cmbBright.SelectedIndex).ToString()
            Else
                Return SEL_STRING
            End If
        End Get
    End Property

    Public ReadOnly Property FadedCluster() As String
        Get
            If Me.cmbFaded.SelectedIndex <> -1 Then
                Return Me.cmbFaded.Items(Me.cmbFaded.SelectedIndex).ToString()
            Else
                Return SEL_STRING
            End If
        End Get
    End Property

    Public ReadOnly Property SymbiantOrSpirit() As String
        Get
            If Me.cmbSymbSpirit.SelectedIndex <> -1 Then
                Return Me.cmbSymbSpirit.Items(Me.cmbSymbSpirit.SelectedIndex).ToString()
            Else
                Return SEL_STRING
            End If
        End Get
    End Property

    Public ReadOnly Property SorI() As String
        Get
            Return Me.lblType.Text
        End Get
    End Property

    Public ReadOnly Property FadedClusterBonus() As Integer
        Get
            Return getClusterBonus(1, Me.FadedCluster(), Me.BrightCluster, Me.ShinyCluster(), prvQL)
        End Get
    End Property

    Public ReadOnly Property BrightClusterBonus() As Integer
        Get
            Return getClusterBonus(2, Me.FadedCluster(), Me.BrightCluster, Me.ShinyCluster(), prvQL)
        End Get
    End Property

    Public ReadOnly Property ShinyClusterBonus() As Integer
        Get
            Return getClusterBonus(3, Me.FadedCluster(), Me.BrightCluster, Me.ShinyCluster(), prvQL)
        End Get
    End Property

    Public ReadOnly Property SymbSpiritBonuses() As String
        Get
            If Me.cmbSymbSpirit.SelectedIndex <> -1 Then
                Return getSymbiantAndSpiritBonuses(Me.cmbSymbSpirit.Items(Me.cmbSymbSpirit.SelectedIndex).ToString())
            Else
                Return ""
            End If
        End Get
    End Property

    Public ReadOnly Property TotalMods() As Integer()
        Get
            Return prvMods
        End Get
    End Property

    Public Property ChangeHandler() As EventHandler
        Get
            Return prvHandler
        End Get
        Set(ByVal Value As EventHandler)
            prvHandler = Value
            prvUseHandler = True
        End Set
    End Property

    Public ReadOnly Property SlotName() As String
        Get
            Dim s_Name As String
            Dim a_Slot() As String = {"Eye", "Head", "Ear", "R-Arm", "Chest", "L-Arm", "R-Wrist", "Waist", "L-Wrist", "R-Hand", "Leg", "L-Hand", "Feet"}

            If Me.ItemIdx < 0 Or Me.ItemIdx > MAX_IMPS Then
                s_Name = "UNKNOWN"
            Else
                s_Name = a_Slot(Me.ItemIdx)
            End If
            Return s_Name
        End Get
    End Property

    Private Sub hideAllControls()
        Me.lblShiny.Visible = False
        Me.lblBright.Visible = False
        Me.lblFaded.Visible = False
        Me.lblQL.Visible = False
        Me.lblSymbSpirit.Visible = False
        Me.lblFixedName.Visible = False
        Me.lblAttrib.Visible = False
        Me.cmbFaded.Visible = False
        Me.cmbBright.Visible = False
        Me.cmbShiny.Visible = False
        Me.nudQL.Visible = False
        Me.cmbSymbSpirit.Visible = False
        ' just disable this
        Me.cmdMaxQL.Enabled = False
    End Sub

    Private Sub showRelevantControls()
        Me.lblShiny.Text = Me.ShinyCluster
        Me.lblBright.Text = Me.BrightCluster
        Me.lblFaded.Text = Me.FadedCluster
        'Me.lblQL.Text = CStr(Me.QL)
        Me.lblSymbSpirit.Text = Me.SymbiantOrSpirit
        If Me.chkLock.Checked Then
            If Me.SorI = "I" Then
                Me.lblShiny.Visible = True
                Me.lblBright.Visible = True
                Me.lblFaded.Visible = True
                Me.lblQL.Visible = True
                Me.lblAttrib.Visible = True
            ElseIf Me.SorI = "S" Then
                Me.lblSymbSpirit.Visible = True
                Me.lblQL.Visible = True
            ElseIf Me.SorI = "A" Then
                Me.lblFixedName.Visible = True
                Me.lblQL.Visible = True
            ElseIf Me.SorI = "C" Then
                Me.lblFixedName.Visible = True
                Me.lblQL.Visible = True
            End If
        Else
            If Me.SorI = "I" Then
                Me.cmbFaded.Visible = True
                Me.cmbBright.Visible = True
                Me.cmbShiny.Visible = True
                Me.nudQL.Visible = True
                Me.lblAttrib.Visible = True
                Me.cmbFaded.Enabled = True
                Me.cmbBright.Enabled = True
                Me.cmbShiny.Enabled = True
                Me.nudQL.Enabled = True
                Me.cmdMaxQL.Enabled = True
            ElseIf Me.SorI = "S" Then
                Me.cmbSymbSpirit.Visible = True
                Me.cmbSymbSpirit.Enabled = True
                Me.lblQL.Visible = True
            ElseIf Me.SorI = "A" Then
                Me.lblFixedName.Visible = True
                Me.nudQL.Visible = True
                Me.nudQL.Enabled = True
                Me.cmdMaxQL.Enabled = True
            ElseIf Me.SorI = "C" Then
                Me.lblFixedName.Visible = True
                Me.nudQL.Visible = True
                Me.nudQL.Enabled = True
                Me.cmdMaxQL.Enabled = True
            End If
        End If
    End Sub

    Public Sub getImplantAOIDs(ByRef lo_aoid As Integer, ByRef hi_aoid As Integer)
        Dim aoid1 As Integer
        Dim aoid2 As Integer
        Dim aoid3 As Integer
        Dim aoid4 As Integer
        Dim strFaded As String
        Dim strBright As String
        Dim strShiny As String
        Dim strSymbSpirit As String

        If Me.SorI = "I" Then
            ' need to convert the loop counter to implant slot number... 2^(i+1)
            ' and skip if all clusters are Empty ?
            strFaded = Me.FadedCluster
            strBright = Me.BrightCluster
            strShiny = Me.ShinyCluster
            If strFaded = SEL_STRING Then
                strFaded = "Empty"
            End If
            If strBright = SEL_STRING Then
                strBright = "Empty"
            End If
            If strShiny = SEL_STRING Then
                strShiny = "Empty"
            End If
            getAOIDsBySlotAndClusters(CInt(2 ^ (Me.ItemIdx + 1)), strFaded, strBright, strShiny, aoid1, aoid2, aoid3, aoid4)
            If Me.QL < 201 Then
                hi_aoid = aoid2
                lo_aoid = aoid1
            Else
                hi_aoid = aoid4
                lo_aoid = aoid3
            End If
        ElseIf Me.SorI = "S" Then
            strSymbSpirit = Me.SymbiantOrSpirit
            ' skip if nothing selected?
            If strSymbSpirit = SEL_STRING Then
                lo_aoid = -1
                hi_aoid = -1
            Else
                ' remove " (QLx)" off end
                strSymbSpirit = strSymbSpirit.Substring(0, strSymbSpirit.IndexOf(" (QL"))
                lo_aoid = symbGetAOIDByName(strSymbSpirit)
                hi_aoid = lo_aoid
            End If
        ElseIf Me.SorI = "A" Then
            ' ideally we'd get these from db...
            lo_aoid = EOE_LO_AOID
            hi_aoid = EOE_HI_AOID
        ElseIf Me.SorI = "C" Then
            ' ideally we'd get these from db...
            lo_aoid = CyborgImplantAOID(0, Me.ItemIdx)
            hi_aoid = CyborgImplantAOID(1, Me.ItemIdx)
        End If
    End Sub

    Public Sub setMaxQL()
        If Me.SorI = "S" Then
            Me.setSymbsAndSpirits(keep_equipped_override:=True)
        ElseIf Me.SorI = "A" Then
            If Me.CheckReqs Then
                Me.MaxQL = determineLimit(getRequirements(EOE_LO_AOID), getQL(EOE_LO_AOID), getRequirements(EOE_HI_AOID), getQL(EOE_HI_AOID))
            Else
                Me.MaxQL = 300
            End If
        ElseIf Me.SorI = "C" Then
            If Me.CheckReqs Then
                Me.MaxQL = determineLimit(getRequirements(CyborgImplantAOID(0, Me.ItemIdx)), getQL(CyborgImplantAOID(0, Me.ItemIdx)), getRequirements(CyborgImplantAOID(1, Me.ItemIdx)), getQL(CyborgImplantAOID(1, Me.ItemIdx)))
            Else
                Me.MaxQL = 200
            End If
        Else
            If Me.CheckReqs Then
                Me.MaxQL = findImplantLimit(Me.ShinyCluster, Me.BrightCluster, Me.FadedCluster, 1)
            Else
                Me.MaxQL = 300
            End If
        End If
    End Sub

    ' will be set up in new handler after initialisation so it doesn't fire too early
    Private Sub chkLock_CheckedChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) ' Handles chkLock.CheckedChanged
        Dim lo_aoid As Integer
        Dim hi_aoid As Integer
        Dim tmpMods(NUM_SKILLS + NUM_ATTRS + NUM_OTHER_STATS + 1) As Integer
        Dim idx As Integer

        If Me.chkLock.Checked Then
            ' if logging on then where the equipped item has changed, log that change
            If Not IsNothing(fmMain.LoggingHandle) Then
                If prvOldType <> Me.SorI Then
                    ' type has changed - record as unequipping old item to empty, then re-equipping new item from empty
                    If prvOldType = "I" Then
                        fmMain.LoggingHandle.LogImplant(Me.SlotName, prvOldShiny, prvOldBright, prvOldFaded, prvOldQL, "Empty", "Empty", "Empty", 0)
                    ElseIf prvOldType = "S" Or prvOldType = "A" Then
                        fmMain.LoggingHandle.LogEquipment(Me.SlotName, prvOldName, prvOldQL, "Empty", 0)
                    End If
                    If Me.SorI = "I" Then
                        fmMain.LoggingHandle.LogImplant(Me.SlotName, "Empty", "Empty", "Empty", 0, Me.ShinyCluster, Me.BrightCluster, Me.FadedCluster, Me.QL)
                    ElseIf Me.SorI = "S" Then
                        fmMain.LoggingHandle.LogEquipment(Me.SlotName, "Empty", 0, Me.SymbiantOrSpirit, Me.QL)
                    ElseIf Me.SorI = "C" Or Me.SorI = "A" Then
                        fmMain.LoggingHandle.LogEquipment(Me.SlotName, "Empty", 0, Me.lblFixedName.Text, Me.QL)
                    End If
                ElseIf prvOldType = "I" Then
                    ' type is Implant, this has not chaged, check each slot and QL
                    If prvOldShiny <> Me.ShinyCluster Or prvOldBright <> Me.BrightCluster Or prvOldFaded <> Me.FadedCluster Or prvOldQL <> Me.QL Then
                        fmMain.LoggingHandle.LogImplant(Me.SlotName, prvOldShiny, prvOldBright, prvOldFaded, prvOldQL, Me.ShinyCluster, Me.BrightCluster, Me.FadedCluster, Me.QL)
                    End If
                ElseIf prvOldType = "S" Then
                    ' type is Symb/Spirit, this has not changed, check Name (this covers QL too as names are QL unique)
                    If prvOldName <> Me.SymbiantOrSpirit Then
                        fmMain.LoggingHandle.LogEquipment(Me.SlotName, prvOldName, prvOldQL, Me.SymbiantOrSpirit, Me.QL)
                    End If
                ElseIf prvOldType = "A" Then
                    ' type is Alien Eye, this has not changed, check QL
                    If prvOldQL <> Me.QL Then
                        fmMain.LoggingHandle.LogEquipment(Me.SlotName, prvOldName, prvOldQL, Me.lblFixedName.Text, Me.QL)
                    End If
                ElseIf prvOldType = "C" Then
                    ' type is Cyborg Implant, this has not changed, check QL
                    If prvOldQL <> Me.QL Then
                        fmMain.LoggingHandle.LogEquipment(Me.SlotName, prvOldName, prvOldQL, Me.lblFixedName.Text, Me.QL)
                    End If
                End If
            End If
            If Not IsNothing(fmMain.LoggingHandle) Then
                fmMain.LoggingHandle.DisableLogging()
            End If
            Me.chkLock.Text = "Ï"
            Me.UseHandler = True
            ' add this item's bonuses to character
            Me.getImplantAOIDs(lo_aoid, hi_aoid)
            ' find the lo and hi aoids and get the ql to populate tmpMods then use that to update the character
            getItemModifiers(lo_aoid, hi_aoid, Me.QL, tmpMods)
            ' process tmpMods as additions
            noRecalc = True
            For i As Integer = 0 To NUM_SKILLS - 1
                skillsCA.Item(i).Implant = skillsCA.Item(i).Implant + tmpMods(i)
            Next i
            For i As Integer = NUM_SKILLS To NUM_SKILLS + NUM_ATTRS - 1
                attrCA.Item(i - NUM_SKILLS).Implant = attrCA.Item(i - NUM_SKILLS).Implant + tmpMods(i)
            Next i
            hpCA.Implant = hpCA.Implant + tmpMods(NUM_SKILLS + NUM_ATTRS)
            npCA.Implant = npCA.Implant + tmpMods(NUM_SKILLS + NUM_ATTRS + 1)
            noRecalc = False
            recalcIP()
            For i As Integer = 0 To NUM_OTHER_STATS - 1
                ' internal ID is 400 + j, find the entry in dspStatArray with that id and assign this mod to it
                idx = statDSP.findById(i + 400)
                If idx <> -1 Then
                    statDSP.Item(idx).StatValue = CStr(CInt(statDSP.Item(idx).StatValue) + tmpMods(i + NUM_SKILLS + NUM_ATTRS + 2))
                End If
            Next
            ' we only need to repop lists etc when an item is unlocked, so nothing further to do here
            If Not IsNothing(fmMain.LoggingHandle) Then
                fmMain.LoggingHandle.EnableLogging()
            End If
            hideAllControls()
            showRelevantControls()
        Else
            ' record the currently selected item (type, name, ql, shiny/bright/faded)
            prvOldType = Me.SorI
            prvOldName = Me.SymbiantOrSpirit
            prvOldQL = Me.QL
            prvOldShiny = Me.ShinyCluster
            prvOldBright = Me.BrightCluster
            prvOldFaded = Me.FadedCluster
            Me.chkLock.Text = "Ð"
            Me.UseHandler = False
            ' remove this item's bonuses from character
            Me.getImplantAOIDs(lo_aoid, hi_aoid)
            ' find the lo and hi aoids and get the ql to populate tmpMods then use that to update the character
            getItemModifiers(lo_aoid, hi_aoid, Me.QL, tmpMods)
            ' process tmpMods as subtractions
            noRecalc = True
            For i As Integer = 0 To NUM_SKILLS - 1
                skillsCA.Item(i).Implant = skillsCA.Item(i).Implant - tmpMods(i)
            Next i
            For i As Integer = NUM_SKILLS To NUM_SKILLS + NUM_ATTRS - 1
                attrCA.Item(i - NUM_SKILLS).Implant = attrCA.Item(i - NUM_SKILLS).Implant - tmpMods(i)
            Next i
            hpCA.Implant = hpCA.Implant - tmpMods(NUM_SKILLS + NUM_ATTRS)
            npCA.Implant = npCA.Implant - tmpMods(NUM_SKILLS + NUM_ATTRS + 1)
            noRecalc = False
            recalcIP()
            For i As Integer = 0 To NUM_OTHER_STATS - 1
                ' internal ID is 400 + j, find the entry in dspStatArray with that id and assign this mod to it
                idx = statDSP.findById(i + 400)
                If idx <> -1 Then
                    statDSP.Item(idx).StatValue = CStr(CInt(statDSP.Item(idx).StatValue) - tmpMods(i + NUM_SKILLS + NUM_ATTRS + 2))
                End If
            Next
            ' recalculate limits on ql for implants and repop the symb/spirit list for this item?
            setMaxQL()
            hideAllControls()
            showRelevantControls()
        End If
    End Sub

    Public Sub loadClusters()
        Dim db_data As SQLite.SQLiteDataReader
        Dim str_faded As String
        Dim str_bright As String
        Dim str_shining As String
        Dim slot_id As Integer

        slot_id = CType(2 ^ (prvItemIdx + 1), Integer)
        db_data = getClustersBySlot(slot_id)
        Me.cmbShiny.Items.Clear()
        Me.cmbShiny.Items.Add(SEL_STRING)
        Me.cmbBright.Items.Clear()
        Me.cmbBright.Items.Add(SEL_STRING)
        Me.cmbFaded.Items.Clear()
        Me.cmbFaded.Items.Add(SEL_STRING)
        Do While db_data.Read()
            str_shining = db_data.GetString(0)
            str_bright = db_data.GetString(1)
            str_faded = db_data.GetString(2)
            If str_shining <> "Empty" Then
                ' add to shiny list
                Me.cmbShiny.Items.Add(str_shining)
            End If
            If str_bright <> "Empty" Then
                ' add to bright list
                Me.cmbBright.Items.Add(str_bright)
            End If
            If str_faded <> "Empty" Then
                ' add to faded list
                Me.cmbFaded.Items.Add(str_faded)
            End If
        Loop
        Me.cmbShiny.SelectedIndex = 0
        Me.cmbBright.SelectedIndex = 0
        Me.cmbFaded.SelectedIndex = 0
    End Sub

    Public Sub clearAll()
        Me.cmbShiny.Items.Clear()
        Me.cmbBright.Items.Clear()
        Me.cmbFaded.Items.Clear()
        Me.cmbSymbSpirit.Items.Clear()
    End Sub

    Public Sub setSymbsAndSpirits(Optional ByVal ignore_override As Boolean = False, Optional ByVal keep_equipped_override As Boolean = False)
        Dim res As String
        Dim resA() As String
        Dim i As Integer
        Dim properName As String
        Dim currentSymb As String
        Dim ignore_limits As Boolean = Not prvCheckReqs
        Dim keep_equipped As Boolean = Me.Lock
        Dim QL As Integer
        Dim maxQL As Integer
        Dim ssAOSInit As aosInitProgress

        If prvProfID = -1 Then
            Return
        End If
        maxQL = 1
        If ignore_override Then
            ignore_limits = True
            keep_equipped = False
        End If
        If ignore_limits Then
            maxQL = 300
        End If
        If keep_equipped_override Then
            keep_equipped = True
        End If
        res = getSymbsAndSpiritsBySlotAndProf(prvItemIdx, prvProfID)
        resA = res.Split("#"c)
        ' record currently selected item (if any) if we are to keep it
        If keep_equipped And Not IsNothing(Me.cmbSymbSpirit.SelectedItem) And Me.SorI = "S" Then
            currentSymb = Me.cmbSymbSpirit.SelectedItem.ToString()
        Else
            currentSymb = "IGNORE"
        End If
        Me.cmbSymbSpirit.Items.Clear()
        Me.cmbSymbSpirit.Items.Add(SEL_STRING)
        ssAOSInit = New aosInitProgress
        ssAOSInit.AOSInitMax = resA.Length - 1
        ssAOSInit.lblText.Text = "Update Symbiant List"
        ssAOSInit.Show(fmImplant)
        ssAOSInit.Refresh()
        For i = 0 To resA.Length - 1
            If resA(i).Length > 0 Then
                ' get proper name and determine if we can use...
                properName = resA(i).Substring(0, resA(i).IndexOf(" (QL", StringComparison.Ordinal))
                If ignore_limits Then
                    Me.cmbSymbSpirit.Items.Add(resA(i))
                Else
                    If isSymbSpiritEquipable(properName) Or resA(i) = currentSymb Then
                        Me.cmbSymbSpirit.Items.Add(resA(i))
                        ' get ql and check if we need to increase the max QL
                        QL = CInt(resA(i).Substring(resA(i).IndexOf(" (QL", StringComparison.Ordinal) + 4, resA(i).Length() - 1 - (resA(i).IndexOf(" (QL", StringComparison.Ordinal) + 4)))
                        If QL > maxQL Then
                            maxQL = QL
                        End If
                    End If
                End If
            End If
            ssAOSInit.AOSInitStep()
            ssAOSInit.Refresh()
        Next
        ssAOSInit.Hide()
        If keep_equipped Then
            Me.cmbSymbSpirit.SelectedItem = currentSymb
        Else
            Me.cmbSymbSpirit.SelectedIndex = 0
        End If
        Me.MaxQL = maxQL
    End Sub

    Public Sub setFocus(ByVal sbf As String)
        Me.Focus()
        If sbf = "Shiny" Then
            Me.cmbShiny.Focus()
        ElseIf sbf = "Bright" Then
            Me.cmbBright.Focus()
        ElseIf sbf = "Faded" Then
            Me.cmbFaded.Focus()
        End If
    End Sub

    Public Sub setElement(ByVal sbf As String, ByVal n As String)
        Dim oldUseHandler As Boolean
        Dim tn As String

        oldUseHandler = prvUseHandler
        prvUseHandler = False
        If sbf = "Shiny" Then
            If Me.lblType.Text = "S" Then
                changeLabel()
                Me.lblSymbSpirit.Text = Me.cmbShiny.Items(0).ToString()
            End If
            If Me.lblType.Text = "A" Or Me.lblType.Text = "C" Then
                changeLabel()
            End If
            If Me.cmbShiny.Items.Contains(n) Then
                Me.cmbShiny.SelectedIndex = Me.cmbShiny.Items.IndexOf(n)
                Me.lblShiny.Text = n
            Else
                Me.cmbShiny.SelectedIndex = 0
                Me.lblShiny.Text = Me.cmbShiny.Items(0).ToString()
            End If
        ElseIf sbf = "Bright" Then
            If Me.lblType.Text = "S" Then
                changeLabel()
                Me.lblSymbSpirit.Text = Me.cmbBright.Items(0).ToString()
            End If
            If Me.lblType.Text = "A" Or Me.lblType.Text = "C" Then
                changeLabel()
            End If
            If Me.cmbBright.Items.Contains(n) Then
                Me.cmbBright.SelectedIndex = Me.cmbBright.Items.IndexOf(n)
                Me.lblBright.Text = n
            Else
                Me.cmbBright.SelectedIndex = 0
                Me.lblBright.Text = Me.cmbBright.Items(0).ToString()
            End If
        ElseIf sbf = "Faded" Then
            If Me.lblType.Text = "S" Then
                changeLabel()
                Me.lblSymbSpirit.Text = Me.cmbFaded.Items(0).ToString()
            End If
            If Me.lblType.Text = "A" Or Me.lblType.Text = "C" Then
                changeLabel()
            End If
            If Me.cmbFaded.Items.Contains(n) Then
                Me.cmbFaded.SelectedIndex = Me.cmbFaded.Items.IndexOf(n)
                Me.lblFaded.Text = n
            Else
                Me.cmbFaded.SelectedIndex = 0
                Me.lblFaded.Text = Me.cmbFaded.Items(0).ToString()
            End If
        ElseIf sbf = "EOE" Then ' Exterminator Occular Enhancement
            If Me.lblType.Text = "I" Then
                changeLabel()
                changeLabel()
                Me.lblShiny.Text = Me.cmbShiny.Items(0).ToString()
                Me.lblBright.Text = Me.cmbBright.Items(0).ToString()
                Me.lblFaded.Text = Me.cmbFaded.Items(0).ToString()
            ElseIf Me.lblType.Text = "S" Then
                changeLabel()
                Me.lblSymbSpirit.Text = Me.cmbShiny.Items(0).ToString()
            End If
        ElseIf sbf = "CI" Then ' Cyborg Implant
            If Me.lblType.Text = "I" Then
                changeLabel()
                changeLabel()
                Me.lblShiny.Text = Me.cmbShiny.Items(0).ToString()
                Me.lblBright.Text = Me.cmbBright.Items(0).ToString()
                Me.lblFaded.Text = Me.cmbFaded.Items(0).ToString()
            ElseIf Me.lblType.Text = "S" Then
                changeLabel()
                Me.lblSymbSpirit.Text = Me.cmbShiny.Items(0).ToString()
            End If
        Else
            If Me.lblType.Text = "I" Then
                changeLabel()
                Me.lblShiny.Text = Me.cmbShiny.Items(0).ToString()
                Me.lblBright.Text = Me.cmbBright.Items(0).ToString()
                Me.lblFaded.Text = Me.cmbFaded.Items(0).ToString()
            End If
            If Me.lblType.Text = "A" Or Me.lblType.Text = "C" Then
                changeLabel()
                changeLabel()
            End If
            ' need to get QL of symbiant in n so we can add " (QLx)" to it if it is missing
            If n.IndexOf(" (QL", StringComparison.Ordinal) = -1 Then
                tn = n & " (QL" & CStr(getQL(symbGetAOIDByName(n))) & ")"
            Else
                tn = n
            End If
            If Me.cmbSymbSpirit.Items.Contains(tn) Then
                Me.cmbSymbSpirit.SelectedIndex = Me.cmbSymbSpirit.Items.IndexOf(tn)
                Me.lblSymbSpirit.Text = tn
            Else
                Me.cmbSymbSpirit.SelectedIndex = 0
                Me.lblSymbSpirit.Text = Me.cmbShiny.Items(0).ToString()
            End If
        End If
        prvUseHandler = oldUseHandler
        If prvUseHandler Then
            prvHandler(Me, Nothing)
            updateMods()
        End If
    End Sub

    Public Function getElement(ByVal sbf As String) As String
        If sbf = "Shiny" Then
            Return Me.ShinyCluster
        ElseIf sbf = "Bright" Then
            Return Me.BrightCluster
        ElseIf sbf = "Faded" Then
            Return Me.FadedCluster
        Else
            Return "ERROR"
        End If
    End Function

    Public Sub Reset()
        Dim oldUseHandler As Boolean

        oldUseHandler = prvUseHandler
        prvUseHandler = False
        If Me.lblType.Text = "S" Then
            changeLabel()
        End If
        If Me.lblType.Text = "A" Then
            changeLabel()
        End If
        If Me.lblType.Text = "C" Then
            changeLabel()
        End If
        Me.QL = 1
        Me.cmbFaded.SelectedIndex = 0
        Me.cmbBright.SelectedIndex = 0
        Me.cmbShiny.SelectedIndex = 0
        Me.ProfessionID = 0
        Me.BreedID = 0
        updateTT()
        prvUseHandler = oldUseHandler
        If prvUseHandler And Not initApp Then
            prvHandler(Me, Nothing)
            updateMods()
        End If
    End Sub

    Private Sub changeLabel()
        If Me.lblType.Text = "I" Then
            If prvProfID = -1 Then
                Return
            End If
            Me.lblType.Text = "S"
            Me.lblType.BackColor = System.Drawing.Color.OliveDrab
            hideAllControls()
            showRelevantControls()
            setSymbsAndSpirits() ' currently assumes false and false as param, we should track the form's chkReqs flag somehow?
            If prvUseHandler Then
                prvHandler(Me, Nothing)
                updateMods()
            End If
        ElseIf Me.lblType.Text = "S" And prvItemIdx = 0 And checkProfAlienEye(prvProfID) Then
            Me.lblType.Text = "A"
            Me.lblType.BackColor = System.Drawing.Color.GreenYellow
            hideAllControls()
            showRelevantControls()
            Me.lblFixedName.Text = "Exterminator Occular Enhancment"
            If prvUseHandler Then
                prvHandler(Me, Nothing)
                updateMods()
            End If
        ElseIf Me.lblType.Text = "S" And checkSlotCyborgImplant(prvItemIdx) And checkBreedCyborgImplant(prvBreedID) Then
            Me.lblType.Text = "C"
            Me.lblType.BackColor = System.Drawing.Color.SandyBrown
            hideAllControls()
            showRelevantControls()
            ' get label of relevant cyborg implant
            Me.lblFixedName.Text = CyborgImplantName(prvItemIdx)
            If prvUseHandler Then
                prvHandler(Me, Nothing)
                updateMods()
            End If
        Else
            Me.lblType.Text = "I"
            Me.lblType.BackColor = System.Drawing.Color.LightSeaGreen
            hideAllControls()
            showRelevantControls()
            If prvUseHandler Then
                prvHandler(Me, Nothing)
                updateMods()
            End If
        End If
        ' update tooltip here
        Me.updateTT()
        setMaxQL()
    End Sub

    Public Sub setType(ByVal t As String)
        If Me.lblType.Text <> t Then
            changeLabel()
            If Me.lblType.Text <> t Then
                changeLabel()
            End If
        End If
    End Sub

    Private Sub lblType_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles lblType.Click
        If Me.chkLock.Checked Then
            Return
        End If
        If prvProfID <> SHADE_ID Then
            changeLabel()
        End If
    End Sub

    Private Sub ihControl1_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        Me.cmbSymbSpirit.Top = 0
        Me.lblQL.Top = 0
        Me.lblFixedName.Top = 0
        Me.lblShiny.Top = 0
        Me.lblBright.Top = 0
        Me.lblFaded.Top = 0
        Me.lblSymbSpirit.Top = 0
        Me.CheckReqs = True
        hideAllControls()
        showRelevantControls()
    End Sub

    Private Sub nudQL_ValueChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles nudQL.ValueChanged
        ' ideally want to supress this unless this control has focus?
        If Me.nudQL Is Me.ActiveControl Then
            prvQL = CType(Me.nudQL.Value, Integer)
            Me.lblQL.Text = CStr(prvQL)
            Me.timeUpdate.Enabled = True
        End If
        hasChanges = True
    End Sub

    Private Sub updateWhenQLChanged(ByVal e As System.EventArgs)
        Me.updateTT()
        If prvUseHandler Then
            prvHandler(Me, e)
            Me.updateMods()
        End If
    End Sub

    Private Sub nudQL_Leave(ByVal sender As Object, ByVal e As System.EventArgs) Handles nudQL.Leave
        If prvQL <> Me.nudQL.Value Then
            prvQL = CType(Me.nudQL.Value, Integer)
            Me.lblQL.Text = CStr(prvQL)
            updateTT()
            If prvUseHandler Then
                prvHandler(Me, e)
                updateMods()
            End If
        End If
    End Sub

    Private Sub cmbSymbSpirit_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmbSymbSpirit.SelectedIndexChanged
        Dim s As String
        Dim p As Integer
        Dim sk As String
        Dim ska() As String
        Dim skv As String
        Dim skk As String
        Dim sc As New SortedList
        Dim de As DictionaryEntry
        Dim j As Integer

        If Me.cmbSymbSpirit.SelectedIndex = -1 Or Me.cmbSymbSpirit.SelectedIndex = 0 Then
            Me.QL = 1
            Return
        End If
        s = Me.cmbSymbSpirit.Items(Me.cmbSymbSpirit.SelectedIndex).ToString()
        p = s.IndexOf("(QL", StringComparison.Ordinal)
        s = s.Substring(p + 3)
        s = s.Substring(0, s.Length - 1)
        Me.QL = CInt(s)
        If Me.SymbiantOrSpirit() <> SEL_STRING Then
            sk = Me.SymbSpiritBonuses()
            ska = sk.Split("#"c)
            For j = 0 To ska.Length - 1
                If ska(j) <> "" Then
                    p = ska(j).IndexOf(" : ", StringComparison.Ordinal)
                    skk = ska(j).Substring(0, p)
                    skv = ska(j).Substring(p + 3)
                    If sc.ContainsKey(skk) Then
                        sc(skk) = CStr(CInt(sc(skk)) + CInt(skv))
                    Else
                        sc.Add(skk, skv)
                    End If
                End If
            Next
            s = ""
            For Each de In sc
                If s <> "" Then
                    s &= Chr(10)
                End If
                s &= convertKeyToName(de.Key.ToString()) & " : " & de.Value.ToString()
            Next
            Me.ttHover.SetToolTip(Me.cmbSymbSpirit, s)
            Me.ttHover.SetToolTip(Me.lblType, s)
            Me.ttHover.SetToolTip(Me.lblSymbSpirit, s)
        Else
            Me.ttHover.SetToolTip(Me.cmbSymbSpirit, Nothing)
            Me.ttHover.SetToolTip(Me.lblType, Nothing)
            Me.ttHover.SetToolTip(Me.lblSymbSpirit, Nothing)
        End If
        If prvUseHandler Then
            prvHandler(Me, e)
            updateMods()
        End If
        hasChanges = True
    End Sub

    Private Sub cmbShiny_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmbShiny.SelectedIndexChanged
        updateTT()
        If prvUseHandler Then
            prvHandler(Me, e)
            updateMods()
        Else
            ' recalc max ql as handler won't
            If Me.CheckReqs Then
                Me.MaxQL = findImplantLimit(Me.ShinyCluster, Me.BrightCluster, Me.FadedCluster, 1)
            Else
                Me.MaxQL = 300
            End If
        End If
        hasChanges = True
    End Sub

    Private Sub cmbBright_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmbBright.SelectedIndexChanged
        updateTT()
        If prvUseHandler Then
            prvHandler(Me, e)
            updateMods()
        Else
            ' recalc max ql as handler won't
            If Me.CheckReqs Then
                Me.MaxQL = findImplantLimit(Me.ShinyCluster, Me.BrightCluster, Me.FadedCluster, 1)
            Else
                Me.MaxQL = 300
            End If
        End If
        hasChanges = True
    End Sub

    Private Sub cmbFaded_SelectedIndexChanged(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmbFaded.SelectedIndexChanged
        updateTT()
        If prvUseHandler Then
            prvHandler(Me, e)
            updateMods()
        Else
            ' recalc max ql as handler won't
            If Me.CheckReqs Then
                Me.MaxQL = findImplantLimit(Me.ShinyCluster, Me.BrightCluster, Me.FadedCluster, 1)
            Else
                Me.MaxQL = 300
            End If
        End If
        hasChanges = True
    End Sub

    Private Sub updateTT()
        Dim s As String = ""
        Dim i As Integer

        If Me.ShinyCluster <> SEL_STRING Then
            s &= Me.ShinyCluster & " : " & CStr(Me.ShinyClusterBonus)
        End If
        If Me.BrightCluster <> SEL_STRING Then
            If s <> "" Then
                s &= Chr(10)
            End If
            s &= Me.BrightCluster & " : " & CStr(Me.BrightClusterBonus)
        End If
        If Me.FadedCluster <> SEL_STRING Then
            If s <> "" Then
                s &= Chr(10)
            End If
            s &= Me.FadedCluster & " : " & CStr(Me.FadedClusterBonus)
        End If
        If s <> "" Then
            Me.ttHover.SetToolTip(Me.cmbShiny, s)
            Me.ttHover.SetToolTip(Me.cmbBright, s)
            Me.ttHover.SetToolTip(Me.cmbFaded, s)
            Me.ttHover.SetToolTip(Me.lblShiny, s)
            Me.ttHover.SetToolTip(Me.lblBright, s)
            Me.ttHover.SetToolTip(Me.lblFaded, s)
            Me.ttHover.SetToolTip(Me.lblType, s)
        Else
            Me.ttHover.SetToolTip(Me.cmbShiny, Nothing)
            Me.ttHover.SetToolTip(Me.cmbBright, Nothing)
            Me.ttHover.SetToolTip(Me.cmbFaded, Nothing)
            Me.ttHover.SetToolTip(Me.lblShiny, Nothing)
            Me.ttHover.SetToolTip(Me.lblBright, Nothing)
            Me.ttHover.SetToolTip(Me.lblFaded, Nothing)
            Me.ttHover.SetToolTip(Me.lblType, Nothing)
        End If
        ' update attribute label here...
        If Me.SorI = "I" Then
            s = getImplantReq(Me.ShinyCluster, Me.BrightCluster, Me.FadedCluster)
            If s = "" Then
                Me.lblAttrib.Text = "None"
            ElseIf s = "ERROR" Then
                Me.lblAttrib.Text = "Error"
            Else
                For i = 0 To 5
                    If abilityFull(i) = s Then
                        s = abilityAbrev(i)
                    End If
                Next
                If s.Length <> 3 Then
                    s = "???"
                End If
                Me.lblAttrib.Text = s
            End If
        ElseIf Me.SorI = "C" Then
            Me.ttHover.SetToolTip(Me.lblFixedName, getCyborgImplantTooltip(CyborgImplantAOID(0, Me.ItemIdx), CyborgImplantAOID(1, Me.ItemIdx), Me.prvQL))
        ElseIf Me.SorI = "A" Then
            Me.ttHover.SetToolTip(Me.lblFixedName, getAlienEyeTooltip(Me.prvQL))
        End If
    End Sub

    Friend Sub updateMods()
        Dim aoid1 As Integer
        Dim aoid2 As Integer
        Dim aoid3 As Integer
        Dim aoid4 As Integer
        Dim strFaded As String = Me.FadedCluster
        Dim strBright As String = Me.BrightCluster
        Dim strShiny As String = Me.ShinyCluster
        Dim strSymbSpirit As String = Me.SymbiantOrSpirit

        'reset array
        For i As Integer = 0 To UBound(prvMods)
            prvMods(i) = 0
        Next
        ' what type of implant is this? get aoid's based on that...
        If Me.lblType.Text = "I" Then
            ' need to convert the loop counter to implant slot number... 2^(i+1)
            ' and skip if all clusters are Empty ?
            If strFaded = SEL_STRING Then
                strFaded = "Empty"
            End If
            If strBright = SEL_STRING Then
                strBright = "Empty"
            End If
            If strShiny = SEL_STRING Then
                strShiny = "Empty"
            End If
            getAOIDsBySlotAndClusters(CInt(2 ^ (Me.ItemIdx + 1)), strFaded, strBright, strShiny, aoid1, aoid2, aoid3, aoid4)
            If Me.QL > 200 Then
                aoid1 = aoid3
                aoid2 = aoid4
            End If
        ElseIf Me.lblType.Text = "S" Then
            ' skip if nothing selected?
            If strSymbSpirit = SEL_STRING Then
                aoid1 = -1
                aoid2 = -1
            Else
                ' remove " (QLx)" off end
                strSymbSpirit = strSymbSpirit.Substring(0, strSymbSpirit.IndexOf(" (QL", StringComparison.Ordinal))
                aoid1 = symbGetAOIDByName(strSymbSpirit)
                aoid2 = aoid1
            End If
        ElseIf Me.lblType.Text = "A" Then
            aoid1 = EOE_LO_AOID
            aoid2 = EOE_HI_AOID
        ElseIf Me.lblType.Text = "C" Then
            aoid1 = CyborgImplantAOID(0, Me.ItemIdx)
            aoid2 = CyborgImplantAOID(1, Me.ItemIdx)
        End If
        If aoid1 <> -1 Then
            getItemModifiers(aoid1, aoid2, prvQL, prvMods)
        End If
    End Sub

    Private Function getImplantReq(ByRef shinyClusterName As String, ByRef brightClusterName As String, ByRef fadedClusterName As String) As String
        Dim result As String
        Dim s_aoids As String
        Dim a_aoids() As String
        Dim lo_aoid As Integer
        'Dim hi_aoid As Integer
        'Dim lo_ql As Integer
        'Dim hi_ql As Integer
        Dim lo_stats(,) As Integer
        'Dim hi_stats(,) As Integer
        Dim attr As Integer
        'Dim i As Integer

        result = ""
        fadedClusterName = IIf(fadedClusterName = SEL_STRING, "Empty", fadedClusterName).ToString()
        brightClusterName = IIf(brightClusterName = SEL_STRING, "Empty", brightClusterName).ToString()
        shinyClusterName = IIf(shinyClusterName = SEL_STRING, "Empty", shinyClusterName).ToString()
        s_aoids = findByClusters(shinyClusterName, brightClusterName, fadedClusterName)
        If s_aoids = "NOMATCH" Then
            Return ""
        Else
            a_aoids = Split(s_aoids, "#"c)
            If prvQL >= 1 And prvQL <= 200 Then
                lo_aoid = CType(a_aoids(0), Integer)
                'hi_aoid = CType(a_aoids(1), Integer)
                'lo_ql = 1
                'hi_ql = 200
            ElseIf prvQL >= 201 And prvQL <= 300 Then
                lo_aoid = CType(a_aoids(2), Integer)
                'hi_aoid = CType(a_aoids(3), Integer)
                'lo_ql = 201
                'hi_ql = 300
            Else
                Return "ERROR"
            End If
            ' get requirement arrays (only need the lo_stats as hi_stats will be the same)
            lo_stats = getRequirements(lo_aoid)
            attr = lo_stats(0, 0)
            If attr = 124 Then
                attr = lo_stats(0, 1)
            End If
            result = objSkillDict.Find_AOSName_From_AOInternalID(attr)
        End If
        Return result
    End Function

    Private Sub timeUpdate_Tick(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles timeUpdate.Tick
        Me.timeUpdate.Enabled = False
        updateWhenQLChanged(e)
    End Sub

    Private Sub cmdMaxQL_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles cmdMaxQL.Click
        ' maximise ql
        Me.QL = Me.MaxQL
        Me.updateTT()
        hasChanges = True
    End Sub

End Class
