'Copyright (c) 2003-2014 Donald Milne
'All rights reserved.
'
'This file is part of AOSkills4.
'
'AOSkills4 is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'AOSkills4 is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with AOSkills4. If not, see <http://www.gnu.org/licenses/>.

Imports System
Imports System.IO
Imports System.Xml
Imports Microsoft.VisualBasic

Module modMain

    Public form_positions(5, 4) As Integer
    Public check_reqs() As Boolean = {True, True}

    Public fmMain As frmMain
    Public fmBuff As frmBuffs
    Public fmBene As frmBenefits
    Public fmPDoll As frmPaperDoll
    Public fmImplant As frmImplants

    Public attrCA As aosSkillAttrItemArray
    Public skillsCA As aosSkillAttrItemArray
    Public hpCA As aosHPNPItem
    Public npCA As aosHPNPItem
    Public statDSP As aosDisplayItemArray

    Public noRecalc As Boolean = False
    Public initApp As Boolean = True ' set to FALSE by frmMain_Load before calling recalcIP() after all controls initialised 
    Public initAppNain As Boolean = True ' set to FALSE by frmMain_VisibleChanged
    Public loadingData As Boolean = False

    Public numPerkCheckboxes As Integer = 0
    Public numAIPerkCheckboxes As Integer = 0
    Public numResearchCheckboxes As Integer = 0

    Public perkCA As aosPerkCheckboxArray
    Public aiperkCA As aosPerkCheckboxArray
    Public leperkCA As aosPerkCheckboxArray

    Public strEquipData As String = ""
    Public strImplantData As String = ""

    Public home_dir As String

    Public impA As aosImplantItemArray

    Public hasChanges As Boolean
    ' these get set to true if we ever open that window, used to save correct position/size info
    Public fmBeneShown As Boolean = False
    Public fmBuffShown As Boolean = False
    Public fmPDollShown As Boolean = False
    Public fmImplantShown As Boolean = False

#If DEBUG Then
    Sub WriteDebug(ByRef info As String)
        Dim fsw_debug As System.IO.StreamWriter

        fsw_debug = New System.IO.StreamWriter(home_dir & Path.DirectorySeparatorChar & "debug.log", True, System.Text.Encoding.UTF8)
        fsw_debug.WriteLine(DateTime.Now.ToString("H:mm:ss.fff") & " : " & info)
        fsw_debug.Close()
    End Sub
#End If

    Public Sub Main()

        home_dir = CurDir()

        initSkillDict(objSkillDict)

        ' attempt to open the item database
        db_conn = openDB(home_dir)
        If IsNothing(db_conn) Then
            MsgBox("Unable to open AOIA item database.", MsgBoxStyle.Critical, "AOSkills4")
            End
        End If

        impA = New aosImplantItemArray

        ' get window positions and sizes
        LoadWindowsData()

        ' create instances of all forms
        fmBuff = New frmBuffs
        fmBene = New frmBenefits
        fmPDoll = New frmPaperDoll
        fmImplant = New frmImplants
        fmMain = New frmMain

        ' load perk data from db (and perks_with_actions.csv file as perk actions are stored as server side hashes)
        LoadPerks()

        ' ensure database has AOS implant table
        fmImplant.initTableIfRequired()

        ' show main form
        fmMain.ShowDialog()

    End Sub

    Public Sub ResetChar()
        Dim i As Integer

        noRecalc = True
        fmMain.nudLevel.Value = 1
        For i = 0 To NUM_SKILLS - 1
            skillsCA.Item(i).Improvement = 0
            skillsCA.Item(i).IPLeft = 1500
        Next
        For i = 0 To NUM_ATTRS - 1
            attrCA.Item(i).Improvement = 0
            attrCA.Item(i).IPLeft = 1500
        Next
        For i = 0 To NUM_OTHER_STATS - 1
            statDSP.Item(i).StatValue = "0"
        Next
        ResetEquipment()
        ClearPerks()
        fmBene.Reset()
        fmBuff.Reset()
        'ClearBuffs()
        noRecalc = False
        recalcIP()
        hasChanges = False
    End Sub

    Public Sub SaveWindowData()
        Dim xwriter As XmlTextWriter

        xwriter = New XmlTextWriter("aoskills_prefs.xml", System.Text.Encoding.UTF8)
        xwriter.Indentation = 4
        xwriter.Formatting = Formatting.Indented
        'Write the beginning of the document including the document declaration. Standalone is true.
        xwriter.WriteStartDocument(True)
        ' save window positions and sizes
        xwriter.WriteStartElement("PREFS")
        xwriter.WriteStartElement("WINDOWS")
        xwriter.WriteStartElement("MAIN")
        xwriter.WriteElementString("TOP", CStr(fmMain.Top))
        xwriter.WriteElementString("LEFT", CStr(fmMain.Left))
        xwriter.WriteElementString("WIDTH", CStr(fmMain.Width))
        xwriter.WriteElementString("HEIGHT", CStr(fmMain.Height))
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("EQUIP")
        If fmPDollShown Then
            xwriter.WriteElementString("TOP", CStr(fmPDoll.Top))
            xwriter.WriteElementString("LEFT", CStr(fmPDoll.Left))
            xwriter.WriteElementString("WIDTH", CStr(fmPDoll.Width))
            xwriter.WriteElementString("HEIGHT", CStr(fmPDoll.Height))
        Else
            xwriter.WriteElementString("TOP", CStr(form_positions(FORM_EQUP, FORM_TOP)))
            xwriter.WriteElementString("LEFT", CStr(form_positions(FORM_EQUP, FORM_LEFT)))
            xwriter.WriteElementString("WIDTH", CStr(form_positions(FORM_EQUP, FORM_WIDTH)))
            xwriter.WriteElementString("HEIGHT", CStr(form_positions(FORM_EQUP, FORM_HEIGHT)))
        End If
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("IMPLANT")
        If fmImplantShown Then
            xwriter.WriteElementString("TOP", CStr(fmImplant.Top))
            xwriter.WriteElementString("LEFT", CStr(fmImplant.Left))
            xwriter.WriteElementString("WIDTH", CStr(fmImplant.Width))
            xwriter.WriteElementString("HEIGHT", CStr(fmImplant.Height))
        Else
            xwriter.WriteElementString("TOP", CStr(form_positions(FORM_IMPLANT, FORM_TOP)))
            xwriter.WriteElementString("LEFT", CStr(form_positions(FORM_IMPLANT, FORM_LEFT)))
            xwriter.WriteElementString("WIDTH", CStr(form_positions(FORM_IMPLANT, FORM_WIDTH)))
            xwriter.WriteElementString("HEIGHT", CStr(form_positions(FORM_IMPLANT, FORM_HEIGHT)))
        End If
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("BUFF")
        If fmBuffShown Then
            xwriter.WriteElementString("TOP", CStr(fmBuff.Top))
            xwriter.WriteElementString("LEFT", CStr(fmBuff.Left))
            xwriter.WriteElementString("WIDTH", CStr(fmBuff.Width))
            xwriter.WriteElementString("HEIGHT", CStr(fmBuff.Height))
        Else
            xwriter.WriteElementString("TOP", CStr(form_positions(FORM_BUFF, FORM_TOP)))
            xwriter.WriteElementString("LEFT", CStr(form_positions(FORM_BUFF, FORM_LEFT)))
            xwriter.WriteElementString("WIDTH", CStr(form_positions(FORM_BUFF, FORM_WIDTH)))
            xwriter.WriteElementString("HEIGHT", CStr(form_positions(FORM_BUFF, FORM_HEIGHT)))
        End If
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("ORG")
        If fmBeneShown Then
            xwriter.WriteElementString("TOP", CStr(fmBene.Top))
            xwriter.WriteElementString("LEFT", CStr(fmBene.Left))
            xwriter.WriteElementString("WIDTH", CStr(fmBene.Width))
            xwriter.WriteElementString("HEIGHT", CStr(fmBene.Height))
        Else
            xwriter.WriteElementString("TOP", CStr(form_positions(FORM_ORG, FORM_TOP)))
            xwriter.WriteElementString("LEFT", CStr(form_positions(FORM_ORG, FORM_LEFT)))
            xwriter.WriteElementString("WIDTH", CStr(form_positions(FORM_ORG, FORM_WIDTH)))
            xwriter.WriteElementString("HEIGHT", CStr(form_positions(FORM_ORG, FORM_HEIGHT)))
        End If
        xwriter.WriteEndElement()
        xwriter.WriteEndElement()
        ' now save check requirements checkbox states
        xwriter.WriteStartElement("CHECKREQS")
        xwriter.WriteElementString("EQUIP", CStr(fmPDoll.chkReqs.Checked))
        xwriter.WriteElementString("IMPLANT", CStr(fmImplant.chkReqs.Checked))
        xwriter.WriteEndElement()
        'End the document
        xwriter.WriteEndElement()
        xwriter.WriteEndDocument()
        'Flush the XML document to the underlying stream and close the underlying stream
        xwriter.Close()

    End Sub

    Public Sub LoadWindowsData()
        Dim rdr As Xml.XmlTextReader
        Dim node_name As String = ""
        Dim window_id As Integer

        ' ensure arry initially 0
        For i As Integer = 0 To form_positions.GetUpperBound(0)
            For j As Integer = 0 To form_positions.GetUpperBound(1)
                form_positions(i, j) = 0
            Next
        Next

        Try
            rdr = New Xml.XmlTextReader("aoskills_prefs.xml")
            Do While rdr.Read()
                Select Case rdr.NodeType
                    Case Xml.XmlNodeType.Element
                        Select Case rdr.Name
                            Case "MAIN"
                                window_id = FORM_MAIN
                            Case "EQUIP"
                                window_id = FORM_EQUP
                            Case "IMPLANT"
                                window_id = FORM_IMPLANT
                            Case "BUFF"
                                window_id = FORM_BUFF
                            Case "ORG"
                                window_id = FORM_ORG
                            Case "CHECKREQS"
                                window_id = -1
                        End Select
                        node_name = rdr.Name
                    Case Xml.XmlNodeType.Text
                        Select Case node_name
                            Case "TOP"
                                form_positions(window_id, FORM_TOP) = CInt(rdr.Value)
                            Case "LEFT"
                                form_positions(window_id, FORM_LEFT) = CInt(rdr.Value)
                            Case "WIDTH"
                                form_positions(window_id, FORM_WIDTH) = CInt(rdr.Value)
                            Case "HEIGHT"
                                form_positions(window_id, FORM_HEIGHT) = CInt(rdr.Value)
                            Case "EQUIP"
                                check_reqs(CHECK_EQUIP) = CBool(rdr.Value)
                            Case "IMPLANT"
                                check_reqs(CHECK_IMPLANT) = CBool(rdr.Value)
                        End Select
                    Case Xml.XmlNodeType.EndElement
                        ' nothing to do here?
                End Select
            Loop
        Catch
            ' if file exists, throw up a warning, otherwise do nothing (we'll create it on exit)
            If System.IO.File.Exists("aoskills_prefs.xml") Then
                MsgBox("There was a problem loading the file aoskills.xml", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
            End If
        Finally
            If Not (rdr Is Nothing) Then
                rdr.Close()
            End If
        End Try
    End Sub

    Public Sub setupOtherStats(TabPage As System.Windows.Forms.TabPage)
        Dim i As Integer
        Dim x As Integer
        Dim y As Integer
        Dim itm As aosDisplayItem

        statDSP = New aosDisplayItemArray
        For i = 0 To UBound(OtherStatList)
            itm = statDSP.AddNewStat(TabPage, OtherStatList(i), CStr(0), OtherStatIDList(i), i)
            ' now place it on the page correctly
            x = CInt(Math.Floor(i / 18)) * (itm.Width + 4) + 4
            y = (i Mod 18) * (itm.Height + 2) + 2
            itm.Left = x
            itm.Top = y
        Next
    End Sub

    Public Function MatchAny(ByVal fixed_value As Integer, ByVal ParamArray values() As Integer) As Boolean
        For index As Integer = 0 To values.Length - 1
            If fixed_value = values(index) Then Return True
        Next
        Return False
    End Function

End Module
