'Copyright (c) 2014 Donald Milne
'All rights reserved.
'
'This file is part of AOSkills4.
'
'AOSkills4 is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'AOSkills4 is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with AOSkills4. If not, see <http://www.gnu.org/licenses/>.

' we use late binding, so this must be off
Option Strict Off

Imports System
Imports System.Windows.Forms
Imports System.Collections.Generic
Imports Microsoft.VisualBasic

Module modPerks

    Private suspendUpdatePerks As Boolean = False

    Public Class Perk

        Private prv_AOID As Integer
        Private prv_Name As String
        Private prv_Number As Integer
        Private prv_Level As Integer
        Private prv_AITitle As Integer
        Private prv_Profs As String
        Private prv_Breeds As String
        Private prv_SpecialName As String
        Private prv_Selected As Boolean

        Public Sub New(ByVal p_aoid As Integer, ByVal p_name As String, ByVal p_num As Integer)
            prv_AOID = p_aoid
            prv_Name = p_name
            prv_Number = p_num
            prv_Level = -1
            prv_Profs = "NNNNNNNNNNNNNN"
            prv_Breeds = "NNNN"
            prv_SpecialName = ""
            prv_Selected = False
            prv_AITitle = 0
        End Sub

        Public ReadOnly Property AOID() As Integer
            Get
                Return prv_AOID
            End Get
        End Property

        Public ReadOnly Property Name() As String
            Get
                Return prv_Name
            End Get
        End Property

        Public ReadOnly Property Number() As Integer
            Get
                Return prv_Number
            End Get
        End Property

        Public Property Level() As Integer
            Set(ByVal Value As Integer)
                prv_Level = Value
            End Set
            Get
                Return prv_Level
            End Get
        End Property

        Public Property AITitle() As Integer
            Set(ByVal Value As Integer)
                prv_AITitle = Value
            End Set
            Get
                Return prv_AITitle
            End Get
        End Property
        ' returns Y or N
        Public Function MatchesProf(ByVal id As Integer) As String
            Return Mid(prv_Profs, id + 1, 1)
        End Function

        ' returns Y or N
        Public Function MatchesBreed(ByVal id As Integer) As String
            Return Mid(prv_Breeds, id + 1, 1)
        End Function

        Public Property Prof() As String
            Set(ByVal Value As String)
                prv_Profs = Value
            End Set
            Get
                Return prv_Profs
            End Get
        End Property

        Public Property Breed() As String
            Set(ByVal Value As String)
                prv_Breeds = Value
            End Set
            Get
                Return prv_Breeds
            End Get
        End Property

        Public Property SpecialName() As String
            Set(ByVal Value As String)
                prv_SpecialName = Value
            End Set
            Get
                Return prv_SpecialName
            End Get
        End Property

        Public Property Selected() As Boolean
            Set(ByVal Value As Boolean)
                prv_Selected = Value
            End Set
            Get
                Return prv_Selected
            End Get
        End Property

    End Class

    Public Class PerkArray : Inherits System.Collections.CollectionBase

        Private prvNumPerks As Integer = 0

        Public Function AddNewPerk(ByVal p_aoid As Integer, ByVal p_name As String, ByVal p_num As Integer) As Perk
            ' handle those duplicated LE perk names where we'd get 1-20 
            ' - this does assume that they are appropriately ordered, but that does seem to be the case for Marksmanship and Intuition which are the only two affected as of 18.7
            Dim aPerk As New Perk(p_aoid, p_name, If(p_num > 10, p_num - 10, p_num))
            Me.List.Add(aPerk)
            Return aPerk
        End Function

        Public Sub Remove()
            ' Check to be sure there is an object to remove.
            If Me.Count > 0 Then
                Me.List.RemoveAt(Me.Count - 1)
            End If
        End Sub

        Default Public ReadOnly Property Item(ByVal Index As Integer) As Perk
            Get
                Return CType(Me.List.Item(Index), Perk)
            End Get
        End Property

        Private Function prv_IndexOfAOID(ByVal p_aoid As Integer) As Integer
            Dim idx As Integer = -1
            For i As Integer = 0 To Me.Count - 1
                If Me.Item(i).AOID = p_aoid Then
                    idx = i
                    Exit For
                End If
            Next
            Return idx
        End Function

        Public ReadOnly Property IndexOfAOID(ByVal p_aoid As Integer) As Integer
            Get
                Return prv_IndexOfAOID(p_aoid)
            End Get
        End Property

        Public Property NumSelectedPerks() As Integer
            Get
                Return prvNumPerks
            End Get
            Set(ByVal Value As Integer)
                prvNumPerks = Value
            End Set
        End Property

    End Class

    Public SLPerks As New PerkArray()
    Public AIPerks As New PerkArray()
    Public LEPerks As New PerkArray()

    Private Sub prv_SetPerks(ByRef perk_array As PerkArray)
        Dim perk_obj As Perk
        Dim reqs As Integer(,)

        For Each perk_obj In perk_array
            ' get reqs from DB for profession [60], breed [4] and level [54]
            reqs = getRequirements(perk_obj.AOID)
            ' now enumerate through reqs array
            For j As Integer = 0 To UBound(reqs, 2)
                If reqs(0, j) = 54 Then
                    ' only 1 level (stored as > X, so add 1)
                    perk_obj.Level = reqs(1, j) + 1
                ElseIf reqs(0, j) = 4 Then
                    ' convert value to a breed position
                    Mid(perk_obj.Breed, reqs(1, j), 1) = "Y"
                ElseIf reqs(0, j) = 60 Then
                    If reqs(1, j) = 0 Then
                        ' some perks are available to particular professions only at lower ranks
                        ' e.g. Counterweight perks #1-3 have Engineer, perks #4-6 have "0"
                        ' so skip those where we get a profession = "0"
                    Else
                        ' convert value to a profession position
                        Mid(perk_obj.Prof, profLookup(reqs(1, j)) + 1, 1) = "Y"
                    End If
                ElseIf reqs(0, j) = 169 Then
                    ' only 1 AI title (stored as > X, so add 1)
                    perk_obj.AITitle = reqs(1, j) + 1
                End If
            Next
            ' now, if Breed = "NNNN", change it to "YYYY"
            If perk_obj.Breed = "NNNN" Then
                perk_obj.Breed = "YYYY"
            End If
            ' now, if Profs = "NNNNNNNNNNNNNN", change it to "YYYYYYYYYYYYYY"
            If perk_obj.Prof = "NNNNNNNNNNNNNN" Then
                perk_obj.Prof = "YYYYYYYYYYYYYY"
            End If
        Next
    End Sub

    Private Sub SetPerks()
        ' set other fields based on database
        prv_SetPerks(SLPerks)
        prv_SetPerks(AIPerks)
        prv_SetPerks(LEPerks)
    End Sub

    Public Sub LoadPerks()
        Dim sr_perks As System.IO.StreamReader
        Dim data_line As String
        Dim split_data() As String
        Dim aoid As Integer
        Dim name As String
        Dim number As Integer
        Dim idx As Integer
        ' declare these as a fixed size
        Dim actions_aoid(999) As Integer
        Dim actions_name(999) As String
        Dim num_perk_actions As Integer = 0
        Dim new_perk As Perk
        Dim db_data As SQLite.SQLiteDataReader
        Dim last_name As String
        Dim exception_list As List(Of Int32)

        ' read the "perks_with_actions.csv" file and attach action names to AOIDs
        sr_perks = New System.IO.StreamReader("perks_with_actions.csv")
        Do
            Try
                data_line = sr_perks.ReadLine()
                ' skip first line as it is the header
                If data_line <> "PerkAOID,ActionName,ActionAOID" Then
                    ' line will be of form : AOID,Name,ActionAOID
                    split_data = data_line.Split(New [Char]() {","c})
                    actions_aoid(num_perk_actions) = CInt(split_data(0))
                    actions_name(num_perk_actions) = split_data(1)
                    num_perk_actions = num_perk_actions + 1
                End If
            Catch e As Exception
                Exit Do
            End Try
        Loop
        sr_perks.Close()
        ' load AOIDs of redundant perks for filtering
        exception_list = LoadExceptionPerks()
        ' now read the actual perks from the db and use above arrays to set any actions
        db_data = getPerkList()
        last_name = ""
        Do While db_data.Read()
            aoid = CInt(db_data("aoid").ToString())
            name = db_data("name").ToString()
            If Not exception_list.Contains(aoid) Then
                ' if name changes, reset perk number
                If name <> last_name Then
                    number = 1
                End If
                ' SL 210830 to 239821 and 281534 to 283682
                ' AI 247748 to 253147 [ATE3 is now 256082]
                ' LE 260870 to 262158 and 303268 to 303277
                If (aoid >= 210830 And aoid <= 239821) Or (aoid >= 281534 And aoid <= 283682) Then
                    new_perk = SLPerks.AddNewPerk(aoid, name, number)
                ElseIf (aoid >= 247748 And aoid <= 253147) Or (aoid = 256082) Then
                    new_perk = AIPerks.AddNewPerk(aoid, name, number)
                ElseIf (aoid >= 260870 And aoid <= 262158) Or (aoid >= 303268 And aoid <= 303277) Then
                    new_perk = LEPerks.AddNewPerk(aoid, name, number)
                Else
                    ' shouldn't be here, so throw an exception?
                    Throw New System.ArgumentOutOfRangeException("Unknown Perk [" & CStr(aoid) & " : " & name & "].")
                End If
                idx = Array.IndexOf(actions_aoid, aoid)
                If idx <> -1 Then
                    new_perk.SpecialName = actions_name(idx)
                End If
                ' increment perk number and store name
                number += 1
                last_name = name
            End If
        Loop
        ' now set remainign values from database
        SetPerks()
    End Sub

    Public Sub AddPerk(ByRef perk_array As PerkArray, ByVal idx As Integer)
        If perk_array(idx).Selected = False Then
            perk_array(idx).Selected = True
            perk_array.NumSelectedPerks += 1
            fmMain.updatePerks()
        End If
    End Sub

    Public Sub RemPerk(ByRef perk_array As PerkArray, ByVal idx As Integer)
        If perk_array(idx).Selected = True Then
            perk_array(idx).Selected = False
            perk_array.NumSelectedPerks -= 1
            fmMain.updatePerks()
        End If
    End Sub

    Public Sub ConfigurePerks()
        Dim old_noRecalc As Boolean

        resetCombinedModArrays()
        calculateTotalModifiers(combinedSkillMods, combinedAttributeMods, combinedHPNPMods, combinedOtherStats)
        old_noRecalc = noRecalc
        noRecalc = True
        ' parse tmpMods and add the bonuses to the relevant stats
        For i As Integer = 0 To NUM_SKILLS - 1
            ' skills
            skillsCA.Item(i).Perk = combinedSkillMods(i, PERK_MODS)
        Next
        For i As Integer = 0 To NUM_ATTRS - 1
            ' attrs
            attrCA.Item(i).Perk = combinedAttributeMods(i, PERK_MODS)
        Next
        hpCA.Perk = combinedHPNPMods(0, PERK_MODS)
        npCA.Perk = combinedHPNPMods(1, PERK_MODS)
        PopulateOtherStats()
        noRecalc = old_noRecalc
        recalcIP()
    End Sub

    Public Function PostIncr(ByRef n As Integer) As Integer
        Dim old_n As Integer
        old_n = n
        n += 1
        Return old_n
    End Function

    Private Function AppendLine(ByVal s1 As String, ByVal s2 As String) As String
        Return CStr(IIf(Len(s1) > 0, s1 & Chr(10) & s2, s2))
    End Function

    Public Function ShowPerk(ByRef perk_array As PerkArray, ByVal idx As Integer) As String
        Dim mods As String = ""
        Dim mods_a() As String
        Dim item_a() As String
        Dim r_string As String = ""

        mods = getStrModifiersByType(perk_array(idx).AOID, MOD_EQUIP)
        ' mods in form "id=val,..." so need to split into arrays and parse, mappping id's to internal names
        mods_a = mods.Split(CChar(","))
        For i As Integer = 0 To mods_a.Length - 1
            If mods_a(i) = "" Then
                ' skip this
            Else
                item_a = mods_a(i).Split(CChar("="))
                ' id = item_a(0)
                ' value = item_a(1)
                ' name = objSkillDict.Find_AOSName_From_AOSID(CInt(item_a(0)))
                r_string = AppendLine(r_string, objSkillDict.Find_AOSName_From_AOSID(CInt(item_a(0))) & " : " & item_a(1))
            End If
        Next
        If perk_array(idx).SpecialName <> "" Then
            r_string = AppendLine(r_string, "[" & perk_array(idx).SpecialName & "]")
        End If
        If r_string = "" Then
            r_string = "Perk has no relevant modifiers"
        End If
        Return r_string
    End Function

    Private Sub prv_ClearPerks(ByRef perk_array As PerkArray)
        Dim perk_obj As Perk

        For Each perk_obj In perk_array
            perk_obj.Selected = False
        Next
        perk_array.NumSelectedPerks = 0
    End Sub

    Public Sub ClearPerks()
        Dim i As Integer
        Dim old_noRecalc As Boolean

        prv_ClearPerks(SLPerks)
        prv_ClearPerks(AIPerks)
        prv_ClearPerks(LEPerks)
        old_noRecalc = noRecalc
        noRecalc = True
        For i = 0 To NUM_ATTRS - 1
            attrCA.Item(i).Perk = 0
        Next
        For i = 0 To NUM_SKILLS - 1
            skillsCA.Item(i).Perk = 0
        Next
        hpCA.Perk = 0
        npCA.Perk = 0
        noRecalc = old_noRecalc
        recalcIP()
        fmMain.updatePerks()
    End Sub

    Public Function GetPerkIndex(ByRef perk_array As PerkArray, ByVal perk_name As String, ByVal perk_num As Integer, Optional ByVal prof_id As Integer = -1) As Integer
        Dim i As Integer

        For i = 0 To perk_array.Count - 1
            If perk_array(i).Name = perk_name And perk_array(i).Number = perk_num Then
                If prof_id = -1 Then
                    Return i
                ElseIf perk_array(i).MatchesProf(prof_id) = "Y" Then
                    Return i
                End If
            End If
        Next
        Return -1
    End Function

    Public Function GetPerksByProfAndBreed(ByRef perk_array As PerkArray, ByVal prof_id As Integer, ByVal breed_id As Integer) As String()
        Dim i As Integer
        Dim perk_list As String = ""

        ' cope with first call when prof combo set...
        If breed_id = -1 Then
            Return Split("", "|")
        End If
        For i = 0 To perk_array.Count - 1
            If perk_array(i).MatchesProf(prof_id) = "Y" And perk_array(i).MatchesBreed(breed_id) = "Y" Then
                If InStr(perk_list, perk_array(i).Name) = 0 Then
                    If Len(perk_list) = 0 Then
                        perk_list = perk_list & perk_array(i).Name
                    Else
                        perk_list = perk_list & "|" & perk_array(i).Name
                    End If
                End If
            End If
        Next
        Return Split(perk_list, "|")
    End Function

    Public Function GetNumberOfEntriesByPerk(ByRef perk_array As PerkArray, ByVal perk As String) As Integer
        Dim i As Integer
        Dim cnt As Integer = 0
        Dim exit_flag As Boolean = False

        For i = 0 To perk_array.Count - 1
            ' need to check profession to handel those LE perks with duplicated names
            If perk_array(i).Name = perk And perk_array(i).MatchesProf(fmMain.ProfIndex) = "Y" Then
                exit_flag = True
                cnt += 1
            Else
                If exit_flag = True Then
                    Exit For
                End If
            End If
        Next
        Return cnt
    End Function

    Private Function GetIndicesByPerk(ByRef perk_array As PerkArray, ByVal perk As String) As Integer()
        Dim i As Integer
        Dim idx() As Integer = {-1, -1, -1, -1, -1, -1, -1, -1, -1, -1}
        Dim idx_cnt As Integer = 0
        Dim exit_flag As Boolean = False

        For i = 0 To perk_array.Count - 1
            ' need to check profession to handel those LE perks with duplicated names
            If perk_array(i).Name = perk And perk_array(i).MatchesProf(fmMain.ProfIndex) = "Y" Then
                exit_flag = True
                idx(PostIncr(idx_cnt)) = i
            Else
                If exit_flag = True Then
                    Exit For
                End If
            End If
        Next
        Return idx
    End Function

    Private Function GetValidPerks(ByRef perk_array As PerkArray, ByVal perk As String, ByVal level As Integer, ByVal prof_id As Integer, ByVal breed_id As Integer, ByVal ai_title As Integer) As Integer()
        Dim i As Integer
        Dim vf() As Integer = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0}
        Dim vf_cnt As Integer = 0
        Dim exit_flag As Boolean = False

        For i = 0 To perk_array.Count - 1
            If perk_array(i).Name = perk Then
                exit_flag = True
                If level >= perk_array(i).Level And perk_array(i).MatchesProf(prof_id) = "Y" And perk_array(i).MatchesBreed(breed_id) = "Y" And ai_title >= perk_array(i).AITitle Then
                    vf(PostIncr(vf_cnt)) = 1
                End If
            Else
                If exit_flag = True Then
                    Exit For
                End If
            End If
        Next
        Return vf
    End Function

    Public Function GetNumberOfSLPerks(ByVal lvl As Integer) As Integer
        If lvl <= 200 Then
            Return CInt(Math.Floor(lvl / 10))
        Else
            Return 20 + lvl - 200
        End If
    End Function

    Private Function GetPerkNamePart(strItem As String) As String
        Dim intPos As Integer
        Dim strName As String

        intPos = strItem.IndexOf(" (", StringComparison.Ordinal)
        If intPos > 0 Then
            strName = strItem.Substring(0, intPos)
        Else
            strName = strItem
        End If
        Return strName
    End Function

    Private Function ConstructPerkListName(strItem As String, intSelected As Integer, intTotal As Integer) As String
        Dim strName As String

        strName = strItem & " (" & CStr(intSelected).Trim() & "/" & CStr(intTotal).Trim() & ")"
        Return strName
    End Function

    Public Function CountSelectedPerks(ByRef perk_array As PerkArray, ByVal perk As String) As Integer
        Dim i As Integer
        Dim idx_cnt As Integer = 0
        Dim exit_flag As Boolean = False

        For i = 0 To perk_array.Count - 1
            ' should possibly check profession here too (as GetNumberOfEntriesByPerk), but as we are explicitly only counting selected perks, we should be fine not doing so
            If perk_array(i).Name = perk Then
                exit_flag = True
                If perk_array(i).Selected Then
                    idx_cnt += 1
                End If
            Else
                If exit_flag = True Then
                    Exit For
                End If
            End If
        Next
        Return idx_cnt
    End Function

    Public Function CountAllSelectedPerks(ByRef perk_array As PerkArray) As Integer
        Dim i As Integer
        Dim idx_cnt As Integer = 0

        For i = 0 To perk_array.Count - 1
            If perk_array(i).Selected Then
                idx_cnt += 1
            End If
        Next
        Return idx_cnt
    End Function

    Public Sub UpdatePerkLabel(ByRef perk_listbox As ListBox, ByRef perk_array As PerkArray, ByVal intIndex As Integer)
        Dim strPerkName As String
        Dim numPerks As Integer
        Dim numSelectedPerks As Integer

        strPerkName = GetPerkNamePart(CStr(perk_listbox.Items(intIndex)))
        numPerks = GetNumberOfEntriesByPerk(perk_array, strPerkName)
        numSelectedPerks = CountSelectedPerks(perk_array, strPerkName)
        strPerkName = ConstructPerkListName(strPerkName, numSelectedPerks, numPerks)
        suspendUpdatePerks = True
        perk_listbox.Items(intIndex) = strPerkName
        suspendUpdatePerks = False
    End Sub

    Private Sub UpdatePerkLabels(ByRef perk_listbox As ListBox, ByRef perk_array As PerkArray)
        Dim intIndex As Integer
        Dim intCount As Integer

        intCount = perk_listbox.Items.Count - 1
        For intIndex = 0 To intCount
            UpdatePerkLabel(perk_listbox, perk_array, intIndex)
        Next
    End Sub

    Public Sub UpdateCheckboxes(ByRef perk_listbox As ListBox, ByRef perk_checkboxes As aosPerkCheckboxArray, ByRef perk_array As PerkArray)
        Dim np As Integer
        Dim ns As Integer
        Dim i As Integer
        Dim j As Integer
        Dim idx() As Integer
        Dim vf() As Integer
        Dim tchk As aosPerkCheckbox
        Dim strName As String

        If suspendUpdatePerks Then
            Return
        End If
        If perk_listbox.SelectedIndex = -1 Then
            If Not (perk_checkboxes Is Nothing) Then
                Do While perk_checkboxes.Count > 0
                    perk_checkboxes.Remove()
                Loop
            End If
            UpdatePerkLabels(perk_listbox, perk_array)
            Return
        End If
        strName = GetPerkNamePart(CStr(perk_listbox.Items(perk_listbox.SelectedIndex)))
        ' extract number of perks from selected perk...
        np = GetNumberOfEntriesByPerk(perk_array, strName)
        ns = CountSelectedPerks(perk_array, strName)
        ' remove existing checkboxes
        fmMain.SuspendLayout()
        For i = 1 To perk_checkboxes.Count
            perk_checkboxes.Remove()
        Next
        idx = GetIndicesByPerk(perk_array, strName)
        vf = GetValidPerks(perk_array, strName, fmMain.NudLevelValue, fmMain.ProfIndex, fmMain.BreedIndex, fmMain.NudAlienTitleValue)
        ' add new checkboxes ( @ 212,28 (first checkbox coors))
        For i = 1 To np
            tchk = perk_checkboxes.AddNewCHK(perk_checkboxes.TabPage, perk_array(idx(i - 1)).Name & " " & CStr(perk_array(idx(i - 1)).Number), idx(i - 1), perk_listbox, perk_array)
            tchk.Top = 24 + ((i - 1) * 26)
            tchk.Left = perk_listbox.Left + perk_listbox.Width + 20
            If vf(i - 1) = 0 Then
                If perk_array(idx(i - 1)).Selected = True Then
                    RemPerk(perk_array, idx(i - 1))
                End If
                tchk.Enabled = False
            Else
                tchk.Enabled = True
            End If
        Next
        numPerkCheckboxes = np
        ' clear checkboxes
        For j = 0 To numPerkCheckboxes - 1
            perk_checkboxes(j).chkControl.Checked = False
        Next
        ' populate values as needed from UserPerks array
        noRecalc = True
        For j = 0 To numPerkCheckboxes - 1
            If perk_array(perk_checkboxes(j).ItemID).Selected = True Then
                perk_checkboxes(j).chkControl.Checked = True
            End If
        Next
        UpdatePerkLabels(perk_listbox, perk_array)
        noRecalc = False
        fmMain.ResumeLayout()
    End Sub

End Module
