'Copyright (c) 2003-2014 Donald Milne
'All rights reserved.
'
'This file is part of AOSkills4.
'
'AOSkills4 is free software: you can redistribute it and/or modify
'it under the terms of the GNU General Public License as published by
'the Free Software Foundation, either version 3 of the License, or
'(at your option) any later version.
'
'AOSkills4 is distributed in the hope that it will be useful,
'but WITHOUT ANY WARRANTY; without even the implied warranty of
'MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
'GNU General Public License for more details.
'
'You should have received a copy of the GNU General Public License
'along with AOSkills4. If not, see <http://www.gnu.org/licenses/>.

Imports System
Imports System.Xml
Imports System.Collections.Generic
Imports Microsoft.VisualBasic

Module modXML

    Private professionMap() As Integer = { _
        -1, 8, -1, -1, -1, 0, 13, -1, -1, 3, 10, -1, -1, -1 _
    }

    Public loadedBuffs As New System.Collections.Specialized.StringCollection

    Private Class WrongProfException : Inherits Exception
    End Class

    Private Class WrongBreedException : Inherits Exception
    End Class

    Private Class LevelTooLowException : Inherits Exception
    End Class

    Private Class AILevelTooLowException : Inherits Exception
    End Class

    Public Class xmlBuffError : Inherits Exception
    End Class

    Public Class BadSkillException : Inherits Exception
    End Class

    Private Function getNamePart(ByVal in_name As String) As String
        Return System.IO.Path.GetFileNameWithoutExtension(in_name)
    End Function

    Public Sub saveDataNew(ByVal filename As String)
        Dim i As Integer
        Dim xwriter As XmlTextWriter
        Dim arrWorking() As String = Nothing
        Dim strSplitString As String = vbCrLf
        Dim caSplitString() As Char = strSplitString.ToCharArray()

        'Create a new XmlTextWriter.
        xwriter = New XmlTextWriter(filename, System.Text.Encoding.UTF8)
        xwriter.Indentation = 4
        xwriter.Formatting = Formatting.Indented
        'Write the beginning of the document including the document declaration. Standalone is true.
        xwriter.WriteStartDocument(True)
        'Write the initial elements
        xwriter.WriteStartElement("CHAR")
        xwriter.WriteElementString("LEVEL", CStr(fmMain.NudLevelValue))
        xwriter.WriteElementString("ALIENTITLE", CStr(fmMain.NudAlienTitleValue))
        xwriter.WriteElementString("PROF", CStr(fmMain.ProfIndex))
        xwriter.WriteElementString("BREED", CStr(fmMain.BreedIndex))
        ' Write Abilities, then Skills
        For i = 0 To NUM_ATTRS - 1
            xwriter.WriteStartElement("ATTR")
            xwriter.WriteAttributeString("ID", CStr(i))
            xwriter.WriteAttributeString("NAME", attrNames(i))
            xwriter.WriteElementString("IP", CStr(attrCA.Item(i).Improvement))
            xwriter.WriteEndElement()
        Next
        For i = 0 To NUM_SKILLS - 1
            xwriter.WriteStartElement("SKILL")
            xwriter.WriteAttributeString("ID", CStr(i))
            xwriter.WriteAttributeString("NAME", skillNames(i))
            xwriter.WriteElementString("IP", CStr(skillsCA.Item(i).Improvement))
            xwriter.WriteEndElement()
        Next
        xwriter.WriteStartElement("PERKS")
        For i = 0 To SLPerks.Count - 1
            ' update to write out in form <PERK NAME="Acrobat" NUM="1" />
            If SLPerks(i).Selected = True Then
                'xwriter.WriteElementString("PERKID", CStr(i))
                xwriter.WriteStartElement("PERK")
                xwriter.WriteAttributeString("NAME", SLPerks(i).Name)
                xwriter.WriteAttributeString("NUM", CStr(SLPerks(i).Number))
                xwriter.WriteEndElement()
            End If
        Next
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("AIPERKS")
        For i = 0 To AIPerks.Count - 1
            ' update to write out in form <AIPERK NAME="Shadow" NUM="1" />
            If AIPerks(i).Selected = True Then
                'xwriter.WriteElementString("AIPERKID", CStr(i))
                xwriter.WriteStartElement("AIPERK")
                xwriter.WriteAttributeString("NAME", AIPerks(i).Name)
                xwriter.WriteAttributeString("NUM", CStr(AIPerks(i).Number))
                xwriter.WriteEndElement()
            End If
        Next
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("RESEARCH")
        For i = 0 To LEPerks.Count - 1
            ' update to write out in form <LEPERK NAME="Assassin's Awareness" NUM="1" />
            If LEPerks(i).Selected = True Then
                'xwriter.WriteElementString("RESEARCHID", CStr(i))
                xwriter.WriteStartElement("LEPERK")
                xwriter.WriteAttributeString("NAME", LEPerks(i).Name)
                xwriter.WriteAttributeString("NUM", CStr(LEPerks(i).Number))
                xwriter.WriteEndElement()
            End If
        Next
        xwriter.WriteEndElement()
        ' write out from fmBuffs.lstSelected.Items
        xwriter.WriteStartElement("BUFFS")
        For i = 0 To fmBuff.lstSelected.Items.Count - 1
            xwriter.WriteElementString("BUFFNAME", CType(fmBuff.lstSelected.Items(i), clsBuffListItem).Name)
        Next
        xwriter.WriteEndElement()
        xwriter.WriteStartElement("BENEFITS")
        ' write out from fmBene.lstSelected.Items
        For i = 0 To fmBene.lstSelected.Items.Count - 1
            xwriter.WriteElementString("BENEFITNAME", CType(fmBene.lstSelected.Items(i), clsOrgBenefitListItem).Name & " @ QL" & CStr(CType(fmBene.lstSelected.Items(i), clsOrgBenefitListItem).QL))
        Next
        xwriter.WriteEndElement()
        ' write equipment info
        xwriter.WriteStartElement("EQUIPMENT")
        For i = 0 To NUM_EQUIP_SLOTS - 1
            xwriter.WriteStartElement("SLOT")
            xwriter.WriteAttributeString("ID", CStr(i))
            xwriter.WriteAttributeString("LO_AOID", CStr(fmPDoll.Equipment.Item(i).LO_AOID))
            xwriter.WriteAttributeString("HI_AOID", CStr(fmPDoll.Equipment.Item(i).HI_AOID))
            xwriter.WriteAttributeString("QL", CStr(fmPDoll.Equipment.Item(i).QL))
            xwriter.WriteEndElement()
        Next
        xwriter.WriteEndElement()
        ' write implant info
        xwriter.WriteStartElement("IMPLANTS")
        For i = 0 To MAX_IMPS
            xwriter.WriteStartElement("IMPLANT")
            xwriter.WriteAttributeString("ID", CStr(i))
            xwriter.WriteAttributeString("QL", CStr(impA(i).QL))
            xwriter.WriteAttributeString("TYPE", impA(i).SorI)
            If impA(i).SorI = "I" Then
                xwriter.WriteAttributeString("SHINY", IIf(impA(i).ShinyCluster = SEL_STRING, "Empty", impA(i).ShinyCluster).ToString())
                xwriter.WriteAttributeString("BRIGHT", IIf(impA(i).BrightCluster = SEL_STRING, "Empty", impA(i).BrightCluster).ToString())
                xwriter.WriteAttributeString("FADED", IIf(impA(i).FadedCluster = SEL_STRING, "Empty", impA(i).FadedCluster).ToString())
            ElseIf impA(i).SorI = "S" Then
                xwriter.WriteAttributeString("SYMB", impA(i).SymbiantOrSpirit)
            ElseIf impA(i).SorI = "A" Then
                ' nothing further to do here?
            End If
            xwriter.WriteEndElement()
        Next
        xwriter.WriteEndElement()
        'End the document
        xwriter.WriteEndElement()
        xwriter.WriteEndDocument()
        'Flush the XML document to the underlying stream and close the underlying stream
        xwriter.Close()
    End Sub

    ' load characters from file
    Public Sub loadDataNew(ByVal file_name As String)
        Dim rdr As Xml.XmlTextReader
        Dim node_name As String = ""
        Dim ver As Integer = 1

        Dim buffs As New Collections.ArrayList
        Dim sl_perks As New Collections.ArrayList
        Dim ai_perks As New Collections.ArrayList
        Dim le_perks As New Collections.ArrayList
        Dim benefits As New Collections.ArrayList

        Dim ssAOSInit As aosInitProgress

        Dim level As Integer = -1
        Dim atitle As Integer = -1
        Dim prof As Integer = -1
        Dim breed As Integer = -1
        Dim id As Integer = -1

        Dim ip As Integer = -1

        Dim ab(NUM_ATTRS) As Integer
        Dim sk(NUM_SKILLS) As Integer

        Dim local_equipment_slots(NUM_EQUIP_SLOTS + MAX_IMPS, 2) As Integer
        Const LES_LO_AOID As Integer = 0
        Const LES_HI_AOID As Integer = 1
        Const LES_QL As Integer = 2

        Dim local_implants(13, 6) As String
        Const IMP_QL As Integer = 0
        Const IMP_TYPE As Integer = 1
        Const IMP_SYMB As Integer = 3
        Const IMP_SHINY As Integer = 4
        Const IMP_BRIGHT As Integer = 5
        Const IMP_FADED As Integer = 6

        Dim local_perks(40, 2) As String
        Dim local_aiperks(30, 2) As String
        Dim local_leperks(80, 2) As String
        Const PERK_NAME As Integer = 0
        Const PERK_NUM As Integer = 1
        Dim perks_count As Integer = 0
        Dim aiperks_count As Integer = 0
        Dim leperks_count As Integer = 0

        Dim noError As Boolean = True

        Dim i As Integer

        Try
            loadingData = True
            rdr = New Xml.XmlTextReader(file_name)
            Do While rdr.Read()
                Select Case rdr.NodeType
                    Case Xml.XmlNodeType.Element
                        Select Case rdr.Name
                            Case "ATTR"
                                id = CInt(rdr.GetAttribute("ID"))
                                ip = 0
                            Case "SKILL"
                                id = CInt(rdr.GetAttribute("ID"))
                                ip = 0
                            Case "SLOT"
                                i = CInt(rdr.GetAttribute("ID"))
                                local_equipment_slots(i, LES_LO_AOID) = CInt(rdr.GetAttribute("LO_AOID"))
                                local_equipment_slots(i, LES_HI_AOID) = CInt(rdr.GetAttribute("HI_AOID"))
                                local_equipment_slots(i, LES_QL) = CInt(rdr.GetAttribute("QL"))
                            Case "IMPLANT"
                                i = CInt(rdr.GetAttribute("ID"))
                                local_implants(i, IMP_QL) = rdr.GetAttribute("QL")
                                local_implants(i, IMP_TYPE) = rdr.GetAttribute("TYPE")
                                If local_implants(i, IMP_TYPE) = "I" Then
                                    local_implants(i, IMP_SHINY) = rdr.GetAttribute("SHINY")
                                    local_implants(i, IMP_BRIGHT) = rdr.GetAttribute("BRIGHT")
                                    local_implants(i, IMP_FADED) = rdr.GetAttribute("FADED")
                                ElseIf local_implants(i, IMP_TYPE) = "S" Then
                                    local_implants(i, IMP_SYMB) = rdr.GetAttribute("SYMB")
                                End If
                            Case "PERK"
                                local_perks(perks_count, PERK_NAME) = rdr.GetAttribute("NAME")
                                local_perks(perks_count, PERK_NUM) = rdr.GetAttribute("NUM")
                                perks_count = perks_count + 1
                            Case "AIPERK"
                                local_aiperks(aiperks_count, PERK_NAME) = rdr.GetAttribute("NAME")
                                local_aiperks(aiperks_count, PERK_NUM) = rdr.GetAttribute("NUM")
                                aiperks_count = aiperks_count + 1
                            Case "LEPERK"
                                local_leperks(leperks_count, PERK_NAME) = rdr.GetAttribute("NAME")
                                local_leperks(leperks_count, PERK_NUM) = rdr.GetAttribute("NUM")
                                leperks_count = leperks_count + 1
                            Case Else
                                node_name = rdr.Name
                        End Select
                    Case Xml.XmlNodeType.Text
                        Select Case node_name
                            Case "LEVEL"
                                level = CInt(rdr.Value)
                            Case "ALIENTITLE"
                                atitle = CInt(rdr.Value)
                            Case "PROF"
                                prof = CInt(rdr.Value)
                            Case "BREED"
                                breed = CInt(rdr.Value)
                            Case "IP"
                                ip = CInt(rdr.Value)
                            Case "BUFFNAME"
                                buffs.Add(rdr.Value)
                            Case "BENEFITNAME"
                                benefits.Add(rdr.Value)
                        End Select
                    Case Xml.XmlNodeType.EndElement
                        Select Case rdr.Name
                            Case "ATTR"
                                ab(id) = ip
                            Case "SKILL"
                                sk(id) = ip
                        End Select
                End Select
            Loop
        Catch
            noError = False
            MsgBox("There was a problem loading the file <" & file_name & ">", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Finally
            If Not (rdr Is Nothing) Then
                rdr.Close()
            End If
            loadingData = False
        End Try
        If noError = True Then
            loadingData = True
            ResetChar()
            noRecalc = True
            ' put values into system
            If level <> -1 Then
                fmMain.NudLevelValue = level
            End If
            If atitle <> -1 Then
                fmMain.NudAlienTitleValue = atitle
            End If
            If prof <> -1 Then
                fmMain.ProfIndex = prof
            End If
            If breed <> -1 Then
                fmMain.BreedIndex = breed
            End If
            ' force update of limits...
            noRecalc = False
            recalcIP()
            noRecalc = True
            ' abilities
            For i = 0 To NUM_ATTRS - 1
                attrCA.Item(i).Improvement = ab(i)
            Next
            ' process buffs
            fmBuff.SetBuffs(buffs)
            ' process perks
            For i = 0 To perks_count - 1
                ' get ID from name/number and AddPerk that
                AddPerk(SLPerks, GetPerkIndex(SLPerks, local_perks(i, PERK_NAME), CInt(local_perks(i, PERK_NUM))))
            Next
            For i = 0 To aiperks_count - 1
                ' get ID from name/number and AddPerk that
                AddPerk(AIPerks, GetPerkIndex(AIPerks, local_aiperks(i, PERK_NAME), CInt(local_aiperks(i, PERK_NUM))))
            Next
            For i = 0 To leperks_count - 1
                ' get ID from name/number and AddPerk that and handle cases where name included the prof name due to duplicate perk names
                If local_leperks(i, PERK_NAME).Contains("(Soldier)") Or local_leperks(i, PERK_NAME).Contains("(MA)") Or local_leperks(i, PERK_NAME).Contains("(Agent)") Then
                    local_leperks(i, PERK_NAME) = local_leperks(i, PERK_NAME).Substring(0, local_leperks(i, PERK_NAME).IndexOf(" ("))
                End If
                AddPerk(LEPerks, GetPerkIndex(LEPerks, local_leperks(i, PERK_NAME), CInt(local_leperks(i, PERK_NUM)), fmMain.ProfIndex))
            Next
            ' process benefits
            fmBene.SetBenefits(benefits)
            ' update perk data
            ConfigurePerks()
            UpdateCheckboxes(fmMain.lstPerks, perkCA, SLPerks)
            UpdateCheckboxes(fmMain.lstAIPerks, aiperkCA, AIPerks)
            UpdateCheckboxes(fmMain.lstResearch, leperkCA, LEPerks)
            ' force update of limits...
            noRecalc = False
            recalcIP()
            noRecalc = True
            ' skills
            For i = 0 To NUM_SKILLS - 1
                skillsCA.Item(i).Improvement = sk(i)
            Next
            ' redo equipment modifiers from the array
            For i = 0 To NUM_EQUIP_SLOTS - 1
                fmPDoll.Equipment.Item(i).SetItem(local_equipment_slots(i, LES_LO_AOID), local_equipment_slots(i, LES_HI_AOID), local_equipment_slots(i, LES_QL))
            Next
            ' setup implants etc
            ssAOSInit = New aosInitProgress
            ssAOSInit.AOSInitMax = impA.Count
            ssAOSInit.lblText.Text = "Update Symbiant List"
            ssAOSInit.Show(fmImplant)
            ssAOSInit.Refresh()
            For i = 0 To MAX_IMPS
                ssAOSInit.AOSInitStep()
                ssAOSInit.Refresh()
                impA(i).UseHandler = False
                impA(i).MaxQL = 300
                impA(i).setType(local_implants(i, IMP_TYPE))
                If impA(i).SorI = "I" Then
                    impA(i).setElement("Shiny", local_implants(i, IMP_SHINY))
                    impA(i).setElement("Bright", local_implants(i, IMP_BRIGHT))
                    impA(i).setElement("Faded", local_implants(i, IMP_FADED))
                    impA(i).MaxQL = CInt(local_implants(i, IMP_QL))
                    impA(i).QL = CInt(local_implants(i, IMP_QL))
                ElseIf impA(i).SorI = "A" Then
                    impA(i).setElement("EOE", "")
                    impA(i).MaxQL = CInt(local_implants(i, IMP_QL))
                    impA(i).QL = CInt(local_implants(i, IMP_QL))
                ElseIf impA(i).SorI = "C" Then
                    impA(i).setElement("CI", "")
                    impA(i).MaxQL = CInt(local_implants(i, IMP_QL))
                    impA(i).QL = CInt(local_implants(i, IMP_QL))
                ElseIf impA(i).SorI = "S" Then
                    impA(i).setSymbsAndSpirits(True)
                    impA(i).setElement("Symbiant", local_implants(i, IMP_SYMB))
                End If
                impA(i).UseHandler = True
                impA(i).updateMods()
            Next
            ssAOSInit.Hide()
            resetCombinedModArrays()
            calculateTotalModifiers(combinedSkillMods, combinedAttributeMods, combinedHPNPMods, combinedOtherStats)
            ' set "Other", "Implant" and "Perk" parts of all skills, abilities etc
            For i = 0 To NUM_SKILLS - 1
                skillsCA.Item(i).Other = combinedSkillMods(i, EQUIP_MODS)
                skillsCA.Item(i).Implant = combinedSkillMods(i, IMP_MODS)
                skillsCA.Item(i).Perk = combinedSkillMods(i, PERK_MODS)
            Next
            For i = 0 To NUM_ATTRS - 1
                attrCA.Item(i).Other = combinedAttributeMods(i, EQUIP_MODS)
                attrCA.Item(i).Implant = combinedAttributeMods(i, IMP_MODS)
                attrCA.Item(i).Perk = combinedAttributeMods(i, PERK_MODS)
            Next
            hpCA.Other = combinedHPNPMods(0, EQUIP_MODS)
            npCA.Other = combinedHPNPMods(1, EQUIP_MODS)
            hpCA.Implant = combinedHPNPMods(0, IMP_MODS)
            npCA.Implant = combinedHPNPMods(1, IMP_MODS)
            hpCA.Perk = combinedHPNPMods(0, PERK_MODS)
            npCA.Perk = combinedHPNPMods(1, PERK_MODS)
            PopulateOtherStats()
            ' final update
            noRecalc = False
            recalcIP()
            loadingData = False
        End If
    End Sub

    ' load characters from profile data
    Public Sub loadRawData(ByVal file_name As String)
        Dim rdr As Xml.XmlTextReader
        Dim node_name As String = ""

        Dim level As Integer = -1
        Dim ailevel As Integer = -1
        Dim prof As Integer = -1
        Dim breed As Integer = -1
        Dim id As Integer = -1

        Dim ip As Integer = -1
        Dim ab(NUM_ATTRS) As Integer
        Dim sk(NUM_SKILLS) As Integer

        Dim noError As Boolean = True

        Dim i As Integer

        Try
            loadingData = True
            rdr = New Xml.XmlTextReader(file_name)
            Do While rdr.Read()
                Select Case rdr.NodeType
                    Case Xml.XmlNodeType.Element
                        Select Case rdr.Name
                            Case "ATTR"
                                id = CInt(rdr.GetAttribute("ID"))
                                ip = 0
                            Case "SKILL"
                                id = CInt(rdr.GetAttribute("ID"))
                                ip = 0
                            Case Else
                                node_name = rdr.Name
                        End Select
                    Case Xml.XmlNodeType.Text
                        Select Case node_name
                            Case "LEVEL"
                                level = CInt(rdr.Value)
                            Case "AILEVEL"
                                ailevel = CInt(rdr.Value)
                            Case "PROF"
                                prof = CInt(rdr.Value)
                            Case "BREED"
                                breed = CInt(rdr.Value)
                            Case "IP"
                                ip = CInt(rdr.Value)
                        End Select
                    Case Xml.XmlNodeType.EndElement
                        Select Case rdr.Name
                            Case "ATTR"
                                ab(id) = ip
                            Case "SKILL"
                                sk(id) = ip
                        End Select
                End Select
            Loop
        Catch
            noError = False
            MsgBox("There was a problem loading the file <" & file_name & ">", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OKOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Finally
            If Not (rdr Is Nothing) Then
                rdr.Close()
            End If
            loadingData = False
        End Try
        If noError = True Then
            loadingData = True
            ResetChar()
            noRecalc = True
            ' put values into system
            If level <> -1 Then
                fmMain.NudLevelValue = level
            End If
            If ailevel <> -1 Then
                fmMain.NudAlienTitleValue = ailevel
            End If
            If prof <> -1 Then
                fmMain.ProfIndex = prof
            End If
            If breed <> -1 Then
                fmMain.BreedIndex = breed
            End If
            ' force update of limits...
            noRecalc = False
            recalcIP()
            noRecalc = True
            ' abilities
            For i = 0 To NUM_ATTRS - 1
                attrCA.Item(i).Improvement = ab(i) - breedInit(breed, i)
            Next
            ' force update of limits...
            noRecalc = False
            recalcIP()
            noRecalc = True
            ' skills
            For i = 0 To NUM_SKILLS - 1
                skillsCA.Item(i).Improvement = sk(i) - BASE_SKILL
            Next
            ' final update
            noRecalc = False
            recalcIP()
            loadingData = False
        End If
    End Sub

    ' save character perks to file
    Public Sub saveData_Perk(ByVal filename As String)
        Dim i As Integer
        Dim xwriter As XmlTextWriter

        'Create a new XmlTextWriter.
        xwriter = New XmlTextWriter(filename, System.Text.Encoding.UTF8)
        xwriter.Indentation = 4
        xwriter.Formatting = Formatting.Indented
        'Write the beginning of the document including the document declaration. Standalone is true.
        xwriter.WriteStartDocument(True)
        'Write the initial elements
        xwriter.WriteStartElement("PERKLIST")
        xwriter.WriteElementString("PROF", CStr(fmMain.ProfIndex))
        ' should maybe only write this if breed specific perks selected...
        xwriter.WriteElementString("BREED", CStr(fmMain.BreedIndex))
        xwriter.WriteElementString("LEVEL", CStr(fmMain.NudLevelValue))
        xwriter.WriteElementString("AILEVEL", CStr(fmMain.NudAlienTitleValue))
        ' write selected perks
        For i = 0 To SLPerks.Count - 1
            If SLPerks(i).Selected = True Then
                xwriter.WriteStartElement("PERK")
                xwriter.WriteAttributeString("NAME", "", SLPerks(i).Name)
                xwriter.WriteString(CStr(SLPerks(i).Number))
                xwriter.WriteEndElement()
            End If
        Next
        For i = 0 To AIPerks.Count - 1
            If AIPerks(i).Selected = True Then
                xwriter.WriteStartElement("AIPERK")
                xwriter.WriteAttributeString("NAME", "", AIPerks(i).Name)
                xwriter.WriteString(CStr(AIPerks(i).Number))
                xwriter.WriteEndElement()
            End If
        Next
        For i = 0 To LEPerks.Count - 1
            If LEPerks(i).Selected = True Then
                xwriter.WriteStartElement("RESEARCH")
                xwriter.WriteAttributeString("NAME", "", LEPerks(i).Name)
                xwriter.WriteString(CStr(LEPerks(i).Number))
                xwriter.WriteEndElement()
            End If
        Next
        'End the document
        xwriter.WriteEndDocument()
        'Flush the XML document to the underlying stream and close the underlying stream
        xwriter.Close()
    End Sub

    ' save character misc to file
    Public Sub saveData_Misc(ByVal filename As String)
        Dim i As Integer
        Dim xwriter As XmlTextWriter

        'Create a new XmlTextWriter.
        xwriter = New XmlTextWriter(filename, System.Text.Encoding.UTF8)
        xwriter.Indentation = 4
        xwriter.Formatting = Formatting.Indented
        'Write the beginning of the document including the document declaration. Standalone is true.
        xwriter.WriteStartDocument(True)
        'Write the initial elements
        xwriter.WriteStartElement("ORGBENEFIT")
        ' write out from fmBene.lstSelected.Items, we will set ID now to 0
        For i = 0 To fmBene.lstSelected.Items.Count - 1
            xwriter.WriteStartElement("ITEM")
            xwriter.WriteAttributeString("QL", CStr(CType(fmBene.lstSelected.Items(i), clsOrgBenefitListItem).QL))
            xwriter.WriteAttributeString("NAME", CType(fmBene.lstSelected.Items(i), clsOrgBenefitListItem).Name)
            xwriter.WriteAttributeString("ID", "0")
            xwriter.WriteEndElement()
        Next
        'End the document
        xwriter.WriteEndElement()
        xwriter.WriteEndDocument()
        'Flush the XML document to the underlying stream and close the underlying stream
        xwriter.Close()
    End Sub

    ' load characters perks from file
    Public Sub loadData_Perk(ByVal filename As String)
        Dim i As Integer
        Dim ab(NUM_ATTRS) As Integer
        Dim sk(NUM_SKILLS) As Integer
        Dim hp As Integer = 0
        Dim np As Integer = 0
        Dim xreader As XmlTextReader
        Dim noError As Boolean = True
        Dim node_name As String = ""
        Dim attr_value As String = ""

        For i = 0 To NUM_ATTRS - 1
            ab(i) = 0
        Next
        For i = 0 To NUM_SKILLS - 1
            sk(i) = 0
        Next

        Try
            'Create a new XmlTextReader.
            xreader = New XmlTextReader(filename)
            Do While xreader.Read()
                If xreader.NodeType = XmlNodeType.Element Then
                    node_name = xreader.Name
                    If xreader.HasAttributes() Then
                        attr_value = xreader.GetAttribute("NAME")
                    End If
                End If
                If xreader.NodeType = XmlNodeType.Text Then
                    If node_name = "PROF" Then
                        If CInt(xreader.Value) <> fmMain.ProfIndex Then
                            Throw New WrongProfException
                        End If
                    End If
                    If node_name = "BREED" Then
                        If CInt(xreader.Value) <> fmMain.BreedIndex Then
                            Throw New WrongBreedException
                        End If
                    End If
                    If node_name = "LEVEL" Then
                        If CInt(xreader.Value) > fmMain.NudLevelValue Then
                            Throw New LevelTooLowException
                        End If
                    End If
                    If node_name = "AILEVEL" Then
                        If CInt(xreader.Value) > fmMain.NudAlienTitleValue Then
                            Throw New AILevelTooLowException
                        End If
                    End If
                    If node_name = "PERK" Then
                        AddPerk(SLPerks, GetPerkIndex(SLPerks, attr_value, CInt(xreader.Value)))
                    End If
                    If node_name = "AIPERK" Then
                        AddPerk(AIPerks, GetPerkIndex(AIPerks, attr_value, CInt(xreader.Value)))
                    End If
                    If node_name = "RESEARCH" Then
                        AddPerk(LEPerks, GetPerkIndex(LEPerks, attr_value, CInt(xreader.Value), fmMain.ProfIndex))
                    End If
                End If
                If xreader.NodeType = XmlNodeType.EndElement Then
                    If xreader.Name = "PERKLIST" Then
                        Exit Do
                    End If
                End If
            Loop
        Catch wpe As WrongProfException
            ' do error thing here
            noError = False
            MsgBox("This perk file is not intended for the currently selected profession", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Catch wpe As WrongBreedException
            ' do error thing here
            noError = False
            MsgBox("This perk file is not intended for the currently selected breed", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Catch
            ' do error thing here
            noError = False
            MsgBox("There was a problem loading the perk file <" & filename & ">", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Finally
            If Not (xreader Is Nothing) Then
                xreader.Close()
            End If
        End Try
        If noError = True Then
            ' put values into system
            ConfigurePerks()
            recalcIP()
        End If
    End Sub

    ' load characters misc from file
    Public Sub loadData_Misc(ByVal filename As String)
        Dim nm As String
        Dim ql As Integer
        Dim xreader As XmlTextReader
        Dim noError As Boolean = True
        Dim benefits As New Collections.ArrayList
        Dim num_benefits As Integer = 0

        Try
            'Create a new XmlTextReader.
            xreader = New XmlTextReader(filename)
            xreader.ReadStartElement("ORGBENEFIT")
            If xreader.EOF = False Then
                Do
                    If xreader.IsStartElement("ITEM") Then
                        ' we now ignore the ID
                        nm = CStr(xreader.GetAttribute("NAME"))
                        ql = CInt(xreader.GetAttribute("QL"))
                        benefits.Add(nm & " @QL" & CStr(ql).Trim())
                        num_benefits += 1
                        xreader.Read()
                    Else
                        Exit Do
                    End If
                Loop
                xreader.ReadEndElement()
            End If
        Catch
            ' do error thing here
            noError = False
            MsgBox("There was a problem loading the org benefit file <" & filename & ">", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Finally
            If Not (xreader Is Nothing) Then
                xreader.Close()
            End If
        End Try
        If noError = True Then
            fmBene.SetBenefits(benefits)
            recalcIP()
        End If
    End Sub

    Public Function LoadBuffsToMove() As Dictionary(Of Int32, Int32)
        Dim rdr As Xml.XmlTextReader
        Dim node_name As String = ""
        Dim nm As String
        Dim new_prof_name As String = ""
        Dim new_prof_id As Int32
        Dim buff_list As New Dictionary(Of Int32, Int32)

        Try
            rdr = New Xml.XmlTextReader("buffs_to_move.xml")
            Do While rdr.Read()
                Select Case rdr.NodeType
                    Case Xml.XmlNodeType.Element
                        node_name = rdr.Name
                        If node_name = "AOID" Then
                            nm = CStr(rdr.GetAttribute("NAME"))
                            new_prof_name = rdr.GetAttribute("MOVETO")
                        End If
                    Case Xml.XmlNodeType.Text
                        Select Case node_name
                            Case "AOID"
                                new_prof_id = Array.IndexOf(Of String)(profNames, new_prof_name)
                                ' need to convert to the in game id?
                                If new_prof_id >= 0 Then
                                    new_prof_id = frmBuffs.profReverseLookup(new_prof_id)
                                    buff_list.Add(CType(rdr.Value, Int32), new_prof_id)
                                Else
                                    MsgBox("Profession not recognised for AOID " & rdr.Value.ToString(), vbOKOnly, "Data Error")
                                End If
                        End Select
                    Case Xml.XmlNodeType.EndElement
                        ' nothing to do here?
                        node_name = ""
                End Select
            Loop
        Catch
            MsgBox("There was a problem loading the file buffs_to_move.xml", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Finally
            If Not (rdr Is Nothing) Then
                rdr.Close()
            End If
        End Try
        Return buff_list
    End Function

    Public Function LoadSpecialBuffs() As List(Of Int32)
        Dim rdr As Xml.XmlTextReader
        Dim node_name As String = ""
        Dim buff_list As New List(Of Int32)

        Try
            rdr = New Xml.XmlTextReader("buffs_to_add.xml")
            Do While rdr.Read()
                Select Case rdr.NodeType
                    Case Xml.XmlNodeType.Element
                        node_name = rdr.Name
                    Case Xml.XmlNodeType.Text
                        Select Case node_name
                            Case "AOID"
                                buff_list.Add(CType(rdr.Value, Int32))
                        End Select
                    Case Xml.XmlNodeType.EndElement
                        ' nothing to do here?
                        node_name = ""
                End Select
            Loop
        Catch
            MsgBox("There was a problem loading the file additional_bufflist.xml", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Finally
            If Not (rdr Is Nothing) Then
                rdr.Close()
            End If
        End Try
        Return buff_list
    End Function

    Public Function LoadExceptionBuffs() As List(Of Int32)
        Dim rdr As Xml.XmlTextReader
        Dim node_name As String = ""
        Dim buff_list As New List(Of Int32)

        Try
            rdr = New Xml.XmlTextReader("buffs_to_remove.xml")
            Do While rdr.Read()
                Select Case rdr.NodeType
                    Case Xml.XmlNodeType.Element
                        node_name = rdr.Name
                    Case Xml.XmlNodeType.Text
                        Select Case node_name
                            Case "AOID"
                                buff_list.Add(CType(rdr.Value, Int32))
                        End Select
                    Case Xml.XmlNodeType.EndElement
                        ' nothing to do here?
                        node_name = ""
                End Select
            Loop
        Catch
            MsgBox("There was a problem loading the file buffs_to_remove.xml", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Finally
            If Not (rdr Is Nothing) Then
                rdr.Close()
            End If
        End Try
        Return buff_list
    End Function

    Public Function LoadExceptionPerks() As List(Of Int32)
        Dim rdr As Xml.XmlTextReader
        Dim node_name As String = ""
        Dim perk_list As New List(Of Int32)

        Try
            rdr = New Xml.XmlTextReader("perks_to_remove.xml")
            Do While rdr.Read()
                Select Case rdr.NodeType
                    Case Xml.XmlNodeType.Element
                        node_name = rdr.Name
                    Case Xml.XmlNodeType.Text
                        Select Case node_name
                            Case "AOID"
                                perk_list.Add(CType(rdr.Value, Int32))
                        End Select
                    Case Xml.XmlNodeType.EndElement
                        ' nothing to do here?
                End Select
            Loop
        Catch
            MsgBox("There was a problem loading the file perks_to_remove.xml", CType(MsgBoxStyle.Exclamation + MsgBoxStyle.OkOnly, Microsoft.VisualBasic.MsgBoxStyle), "File Open Error")
        Finally
            If Not (rdr Is Nothing) Then
                rdr.Close()
            End If
        End Try
        Return perk_list
    End Function

End Module
